import os
import sys
import unittest

from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetInfo

cur_path = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(os.path.abspath(os.path.join(cur_path, '../')))

from modelbest_sdk.dataset.prefetch_chunk_dataset import PrefetchChunkDataset
from modelbest_sdk.dataset.segment.segment_factory import CONDITIONAL_FIXED_LENGTH_SEGMENT, NO_SEGMENT, FIXED_LENGTH_SEGMENT
from modelbest_sdk.dataset.segment_dataset import SegmentDataset
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder
from test.test_base import TestBase


class TestSegment(TestBase):
    def test_no_segment(self):
        TestBase.generate_data(dataset_path='/tmp/no_segment', file_count=1, doc_count_per_file=10, token_count=10, token_offset=1000, mask_pattern='text')
        context = DatasetContext()
        dataset_info = DatasetInfo('/tmp/no_segment', max_epoch=1)
        p_dataset = PrefetchChunkDataset(context, dataset_info)
        dataset = SegmentDataset(
            p_dataset,
            segment_type=NO_SEGMENT,
            max_len=16
        )
        for data in dataset:
            base_doc = data.base_doc
            expected_token_ids = list(range(1000, 1010))
            assert base_doc.token_ids == expected_token_ids, f"Token IDs do not match. Expected: {expected_token_ids}, Got: {base_doc.token_ids}"
            expected_mask = [False] * 9 + [True]
            assert base_doc.mask == expected_mask, f"Mask does not match. Expected: {expected_mask}, Got: {base_doc.mask}"
    
    def test_fixed_length_segment(self):
        max_len = 16
        token_count = 10
        TestBase.generate_data(dataset_path='/tmp/fixed_length_segment', file_count=1, doc_count_per_file=10, token_count=token_count, token_offset=1000, mask_pattern='text')
        context = DatasetContext()
        dataset_info = DatasetInfo('/tmp/fixed_length_segment', max_epoch=1)
        p_dataset = PrefetchChunkDataset(context, dataset_info)
        dataset = SegmentDataset(
            p_dataset,
            segment_type=FIXED_LENGTH_SEGMENT,
            max_len=16
        )
        start_idx = 0
        expected_token_ids = list(range(1000, 1000+token_count))
        expected_mask = [False] * (token_count - 1) + [True]
        concatenated_token_ids = expected_token_ids * 10
        concatenated_mask = expected_mask * 10
        for data in dataset:
            token_ids = data.base_doc.token_ids
            mask = data.base_doc.mask
            assert len(token_ids) == max_len + 1, f"Token IDs length does not match. Expected: {max_len + 1}, Got: {len(token_ids)}"
            assert len(mask) == max_len + 1, f"Mask length does not match. Expected: {max_len + 1}, Got: {len(mask)}"
            assert token_ids == concatenated_token_ids[start_idx:start_idx + max_len + 1], f"Token IDs do not match. Expected: {concatenated_token_ids[start_idx:start_idx + max_len + 1]}, Got: {token_ids}"
            assert mask == concatenated_mask[start_idx:start_idx + max_len + 1], f"Mask does not match. Expected: {concatenated_mask[start_idx:start_idx + max_len + 1]}, Got: {mask}"
            start_idx += max_len
            
    def test_conditional_fixed_length_segment(self):
        max_len = 16
        token_count = 5 
        # 5, 10, 15, and the last token could not fit into the fixed length segment
        TestBase.generate_data(dataset_path='/tmp/conditional_fixed_length_segment', file_count=1, doc_count_per_file=10, token_count=token_count, token_offset=1000, mask_pattern='instruction')
        context = DatasetContext()
        dataset_info = DatasetInfo('/tmp/conditional_fixed_length_segment', max_epoch=1)
        p_dataset = PrefetchChunkDataset(context, dataset_info)
        dataset = SegmentDataset(
            p_dataset,
            segment_type=CONDITIONAL_FIXED_LENGTH_SEGMENT,
            max_len=16
        )
        expected_token_ids = list(range(1000, 1005)) * 3
        expected_mask = ([True] * 2 + [False] * 2 + [True]) * 3
        for data in dataset:
            token_ids = data.base_doc.token_ids
            mask = data.base_doc.mask
            assert len(token_ids) == token_count * 3, f"Token IDs length does not match. Expected: {token_count * 3}, Got: {len(token_ids)}"
            assert len(mask) == token_count * 3, f"Mask length does not match. Expected: {token_count * 3}, Got: {len(mask)}"
            assert token_ids == expected_token_ids, f"Token IDs do not match. Expected: {expected_token_ids}, Got: {token_ids}"
            assert mask == expected_mask, f"Mask does not match. Expected: {expected_mask}, Got: {mask}"


if __name__ == '__main__':
    unittest.main()




from collections import defaultdict, deque
import hashlib
from typing import Dict, Generator, List, Tuple

import numpy as np
import torch
from modelbest_sdk.dataset.batch_packer.batch_packer import BatchPacker
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk


class CpmFlashAttnBatchPacker(BatchPacker):
    def __init__(self, batch_size: int, max_len: int, dataset_cnt: int, **kwargs):
        super().__init__(batch_size, max_len, dataset_cnt)
        self.buffer = deque()
        self.current_length = 0
        self.max_total_length = batch_size * max_len
        self.batch_size = 1
        self.dataset_cnt = dataset_cnt

    def put(self, data: DetailedDoc):
        self.buffer.append(data)
        self.current_length += len(data.base_doc.token_ids)
        
    def pop(self) -> DetailedDoc:
        lengths = []
        indexes: List[Tuple[int, Dict[Chunk, List[int]]]] = []
        inputs = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        targets = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-100)
        dataset_ids = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-1)
        position_ids = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        tags = torch.full((self.batch_size, self.max_total_length), dtype=torch.int64, fill_value=-1)

        span_begin = 0
        while self.buffer:
            data: DetailedDoc = self.buffer.pop()
            dataset_idx = data.dataset_idx
            doc = data.base_doc
            token_ids = doc.token_ids
            mask = doc.mask
            tag = doc.tag[0] # TODO: support multiple tags
            target_ids = np.where(mask[1:], -100, token_ids[1:]).tolist() + [-100]
            span_end = span_begin + len(token_ids)
            # TODO: what if the inputs is longer than max_total_length?
            # RuntimeError: The expanded size of the tensor (16) must match the existing size (385) at non-singleton dimension 0.  Target sizes: [16].  Tensor sizes: [385]
            inputs[0, span_begin:span_end] = torch.tensor(token_ids, dtype=torch.int32)
            targets[0, span_begin:span_end] = torch.tensor(target_ids, dtype=torch.int32)
            dataset_ids[0, span_begin:span_end] = torch.tensor(dataset_idx, dtype=torch.int32)
            position_ids[0, span_begin:span_end] = torch.from_numpy(np.arange(len(token_ids), dtype=np.int32))
            lengths.append(len(token_ids))
            indexes.append((dataset_idx, data.indexes_dict))
            tags[0, span_begin:span_end] = self.encode_tags(len(token_ids), tag)
            span_begin = span_end
        cu_seqlens = torch.cat(
            [torch.tensor([0] + lengths).cumsum(dim=-1), torch.tensor([self.max_total_length], dtype=torch.int32)],
            dim=0,
        ).int()
        batch = {
            "input_ids": inputs,
            "target_ids": targets,
            "dataset_ids": dataset_ids,
            "indexes": indexes,
            "cu_seqlens": cu_seqlens,
            "max_seqlen": int(torch.max(cu_seqlens[1:] - cu_seqlens[:-1])),
            "lengths": torch.tensor(sum(lengths)).int(),
            "position_ids": position_ids,
            "tags": tags,
            "hash_to_tag": self.hash_to_tag
        }
        self.current_length = 0
        return batch
    
    
    def will_exceed(self, data: DetailedDoc):
        if data is None:
            return False
        return self.current_length + len(data.base_doc.token_ids) > self.max_total_length

    @staticmethod
    def collate_fn(batch):
        return batch[0]

    def __call__(self, detailed_doc: DetailedDoc=None, pop_last=False) -> Generator[DetailedDoc, None, None]:
        if (pop_last and self.buffer):
            yield self.pop()
        if detailed_doc is not None:
            if self.will_exceed(detailed_doc):
                yield self.pop()
            self.put(detailed_doc)



import numpy as np

from modelbest_sdk.dataset.sampler.sampler import Sampler


class WeightedMegatronSampler(Sampler):
    def __init__(self, weights, rank, world_size, seed, **kwargs):
        self.weights = np.array(weights)
        self.rank = rank
        self.world_size = world_size
        self.sample_idx = 0
        self.current_samples = np.zeros(len(self.weights), dtype=np.int64)
        self.kwargs = kwargs
        self.generator = self.generate_indices()
        for _ in range(self.rank):
            next(self.generator)
                    
    def generate_indices(self):
        while True:
            errors = self.weights * self.sample_idx - self.current_samples
            max_error_index = np.argmax(errors)
            self.current_samples[max_error_index] += 1
            yield max_error_index
            self.sample_idx += 1

    def __call__(self):
        index = next(self.generator)
        for _ in range(self.world_size - 1):
            next(self.generator)
        return index
    
    def resume(self, sample_idx, current_samples):
        self.sample_idx = sample_idx
        self.current_samples = current_samples
        self.generator = self.generate_indices()
        for _ in range(self.rank):
            next(self.generator)
    
if __name__ == '__main__':
    weights = [0.39, 0.11, 0.05, 0.08, 0.05, 0.02, 0.24, 0.06]
    rank = 15
    world_size = 16
    
    sampler = WeightedMegatronSampler(weights, rank, world_size)
    for i in range(30):
        print(sampler())


from collections import deque
from typing import Generator
from modelbest_sdk.dataset.segment.segment import Segment
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import LastSample


class FixedLengthSegment(Segment):
    def __init__(self, max_len: int, drop_last=False, **kwargs):
        self.max_len = max_len
        if kwargs.get('custum_segment_length') is not None:
            self.segment_length = kwargs['custum_segment_length']
        else:
            self.segment_length = self.max_len + 1
        self.drop_last = drop_last
        
        self.buffer = deque()
        self.current_length = 0
        
    def put(self, data: DetailedDoc):
        self.buffer.append(data)
        self.current_length += len(data)

    def pop(self) -> DetailedDoc:
        remaining_seq_length = self.segment_length
        begin = 0
        concat_data = None
        last_sample = None
        while self.buffer and remaining_seq_length > 0:
            data: DetailedDoc = self.buffer.popleft()
            self.current_length -= len(data)
            
            if remaining_seq_length >= len(data):
                end = begin + len(data)
                data_to_fill = data
            else:
                end = begin + remaining_seq_length
                data_to_fill, data_to_buffer = data.split(remaining_seq_length)
                self.buffer.appendleft(data_to_buffer)
                self.current_length += len(data_to_buffer)
                # useful when resume
                lask_chunk = data_to_fill.position.chunk
                last_index = data_to_fill.position.index
                last_offset = data_to_fill.prev_len + remaining_seq_length
                last_sample = LastSample(lask_chunk, last_index, last_offset)
            concat_data = data_to_fill if concat_data is None else concat_data.concat(data_to_fill)
            remaining_seq_length -= len(data_to_fill)
            begin = end
        concat_data.last_sample = last_sample
        return concat_data
            
            
    def __call__(self, detailed_doc: DetailedDoc) -> Generator[DetailedDoc, None, None]:
        self.put(detailed_doc)
        while self.current_length > self.max_len:
            yield self.pop()




import json
import os
import unittest
import sys
import thriftpy2
from thriftpy2.utils import deserialize, serialize

cur_path = os.path.realpath(os.path.dirname(__file__))
sys.path.append(os.path.abspath(os.path.join(cur_path, "../")))

from modelbest_sdk.dataset.thrift_wrapper.metadata import SSTableMetadata
from modelbest_sdk.file_format.mbtable import MbTable, MbTableListIterator
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder

metadata_thrift = thriftpy2.load('modelbest_sdk/proto/metadata.thrift')


class TestListTable(unittest.TestCase):
    def setUp(self):
        self.path = "test/data/list_table.mbt"
        builder = MbTableBuilder(self.path)
        for i in range(10):
            values=[]
            for j in range(9):
                values.append(json.dumps({"col": f"value_{i}_{j}"}))
            values.append("")
            builder.write_list(values)
        meta = metadata_thrift.SSTableMetadata(
                tabletype=metadata_thrift.TableType.LIST_TABLE,
                total_count=10)
        builder.add_metadata('metadata', serialize(meta))
        builder.add_metadata('other', 'value')
        builder.flush()
        self.mbtable = MbTable(self.path)
    
    def tearDown(self):
        os.remove(self.path)
    
    def test_get_metadata(self):
        meta = SSTableMetadata.deserialize(self.mbtable.get_metadata('metadata'))
        self.assertEqual('LIST_TABLE', meta.table_type)
        self.assertEqual(10, meta.total_count)
        self.assertEqual(b'value', self.mbtable.get_metadata('other'))
        self.assertEqual(b'', self.mbtable.get_metadata('not_exist'))

        meta = SSTableMetadata.deserialize(self.mbtable.get_file_metadata('metadata'))
        self.assertEqual('LIST_TABLE', meta.table_type)
        self.assertEqual(10, meta.total_count)
        self.assertEqual(b'value', self.mbtable.get_file_metadata('other'))
        self.assertEqual(b'', self.mbtable.get_file_metadata('not_exist'))
        
        meta = SSTableMetadata.deserialize(self.mbtable.get_file_metadata('non_exist'))
        self.assertEqual(None, meta.table_type)
        self.assertEqual(None, meta.total_count)

    def test_get_entry_count(self):
        self.assertEqual(self.mbtable.get_entry_count(), 100)
        self.assertEqual(self.mbtable.get_file_entry_count(), 100)
        self.assertEqual(10, self.mbtable.get_entry_count_from_metadata())

    def test_get_record(self):
        with self.assertRaises(TypeError):
            self.mbtable.read(0)

    def test_get_list(self):
        for i in range(10): 
            values=[]
            for j in range(9):
                values.append(json.dumps({"col": f"value_{i}_{j}"}).encode())
            values.append(b"")
            self.assertEqual(values, self.mbtable.read_list(i)) 


class TestMbTableListIterator(TestListTable):
    def test_iterator(self):
        cnt = 0
        with MbTableListIterator(self.path) as iterator:
            expected = []
            for i in range(10):
                values = []
                for j in range(9):
                    values.append(json.dumps({"col": f"value_{i}_{j}"}).encode())
                values.append(b"")
                expected.append(values)
            for record in iterator:
                self.assertEqual(expected[cnt], record)
                cnt += 1
        assert cnt == 10
    
    def test_iterator_raw(self):
        iterator = MbTableListIterator(self.path)
        iterator.__enter__()
        record = iterator.__next__()
        iterator.__exit__(None, None, None)
        expected =[]
        for j in range(9):
            expected.append(json.dumps({"col": f"value_0_{j}"}).encode())
        expected.append(b"")
        self.assertEqual(expected, record)
        pass
        

    def test_iterator_with_args(self):
        key, cnt = 5, 0
        with MbTableListIterator(self.path, start_index=key, max_iterations=2) as iterator:  
            for _ in iterator:
                cnt += 1
        assert cnt == 2
        
        key, cnt = 9, 0
        with MbTableListIterator(self.path, start_index=key, max_iterations=1000) as iterator:  
            for _ in iterator:
                cnt += 1
        assert cnt == 1
    
        key, cnt = 1000, 0
        # !!! note that key in mbtable is string and is sorted in lexicographic order
        # so key 1000 is greater than 10 but less than 11, this would start from 11
        with MbTableListIterator(self.path, start_index=key, max_iterations=1000) as iterator:  
            for _ in iterator:
                cnt += 1
        assert cnt == 0


if __name__ == '__main__':
    unittest.main()



from typing import Dict, List
import torch

from modelbest_sdk.dataset.prefetch_chunk_dataset import PrefetchChunkDataset
from modelbest_sdk.dataset.segment.segment_factory import FIXED_LENGTH_SEGMENT, NO_SEGMENT, SegmentFactory
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk, DatasetCheckpoint, LastSample
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class SegmentDataset(torch.utils.data.IterableDataset):
    def __init__(
        self, 
        prefetch_chunk_dataset: PrefetchChunkDataset, 
        segment_type=NO_SEGMENT, 
        max_len=1024,
        **kwargs
        ):
        self.name = prefetch_chunk_dataset.path.split('/')[-1]
        self.prefetch_chunk_dataset = prefetch_chunk_dataset
        
        self.segment = SegmentFactory.create_segment(segment_type, max_len, drop_last=False, **kwargs)
        self.load_last_sample = None
        self.last_sample = None

    def __iter__(self):
        for detailed_doc in self.prefetch_chunk_dataset:
            detailed_doc: DetailedDoc
            if not detailed_doc.deserialize():
                continue
            detailed_doc.tag_or_default([self.name])
            if self.load_last_sample is not None and detailed_doc.position.chunk == self.load_last_sample.chunk and detailed_doc.position.index == self.load_last_sample.index:
                _, detailed_doc = detailed_doc.split(self.load_last_sample.offset)
                self.load_last_sample = None
            for data in self.segment(detailed_doc):
                if data is not None:
                    yield data 
    
    def checkpoint(self):
        ckpt = self.prefetch_chunk_dataset.checkpoint()
        ckpt.last_sample = self.last_sample
        return ckpt
    
    def load_checkpoint(self, checkpoint: DatasetCheckpoint):
        self.prefetch_chunk_dataset.load_checkpoint(checkpoint)
        self.load_last_sample = checkpoint.last_sample
        self.last_sample = self.load_last_sample
            
    def update(self, consumed_sample_indexes: Dict[Chunk, List[int]], last_sample: LastSample):
        self.prefetch_chunk_dataset.update(consumed_sample_indexes)
        self.last_sample = last_sample
    
    def __len__(self):
        return len(self.prefetch_chunk_dataset)
    
    
                
if __name__ == '__main__':

    dataset = SegmentDataset(
        context=DatasetContext(),
        path='human.BaseDoc.sstable',
        max_epoch=1,
        segment_type=FIXED_LENGTH_SEGMENT,
        max_len=16
    )
    i = 0
    for data in dataset:
        print(data)
        i += 1
        if i == 100:
            break


import logging
import os
from typing import Dict, List
import torch
from modelbest_sdk.dataset.batch_packer.batch_packer_factory import CPM_FLASH_ATTN_BATCH_PACKER
from modelbest_sdk.dataset.batched_dataset import BatchedDataset
from modelbest_sdk.dataset.sampler.sampler_factory import *
from modelbest_sdk.dataset.segment.segment_factory import *
from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetCheckpointList, DatasetInfoList
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset

logger = logging.getLogger(__name__)
class ModelbestDataloader():
    def __init__(
        self,
        context: DatasetContext,
        dataset_info_list: DatasetInfoList,
        batch_size=1,
        max_len=4096,
        prefetch_chunk_cnt=2,
        chunk_size=1024,
        num_workers=1,
        prefetch_factor=4,
        cuda_prefetch=False,
        segment_type=NO_SEGMENT,
        sampler_type=WEIGHTED_SAMPLER,
        batch_packer_type=CPM_FLASH_ATTN_BATCH_PACKER,
        dp_group=None,
        **kwargs
    ):
        self.context = context
        self.context.num_workers = num_workers
        
        self.weighted_dataset = WeightedDataset(
            context=context, 
            dataset_info_list=dataset_info_list,
            segment_type=segment_type,
            sampler_type=sampler_type,
            max_len=max_len,
            prefetch_chunk_cnt=prefetch_chunk_cnt,
            chunk_size=chunk_size,
            dp_group=dp_group,
            **kwargs
        )
        
        self.batched_dataset = BatchedDataset(
            context=context, 
            weighted_dataset=self.weighted_dataset, 
            batch_size=batch_size, 
            max_len=max_len,
            batch_packer_type=batch_packer_type,
            drop_last=False,
            **kwargs
        )
        
        self.batch_size = 1 if batch_packer_type == CPM_FLASH_ATTN_BATCH_PACKER else batch_size
        self.dataloader = torch.utils.data.DataLoader(
            dataset=self.batched_dataset, 
            batch_size=self.batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=True,
            collate_fn=self.batched_dataset.collate_fn,
            # worker_init_fn=self.weighted_dataset.worker_init,
        )
        
        self.cuda_prefetch = cuda_prefetch
        
        if self.cuda_prefetch:
            from modelbest_sdk.dataset.cuda_prefetcher import CudaPrefetcher
            self.cuda_prefetcher = CudaPrefetcher(
                context,
                self.dataloader
            )
        self.iterator = None

    def __iter__(self):
        if not self.cuda_prefetch:
            self.iterator = iter(self.dataloader)
        else:
            self.iterator = iter(self.cuda_prefetcher)
        return self
    
    def __next__(self):
        if self.iterator is None:
            self.__iter__()
        return next(self.iterator)
    
    def checkpoint(self):
        return self.weighted_dataset.checkpoint()
    
    def load_checkpoint(self, checkpoint):
        self.weighted_dataset.load_checkpoint(checkpoint)
        
    def update(self, dataset_entries, last_samples={}):
        self.weighted_dataset.update(dataset_entries, last_samples)
    
    def save(self, sub_dir=''):
        if self.context.tp_rank != 0 or self.context.pp_rank != 0:
            logger.warning("Only tp/pp rank 0 can save dataloader")
            return
        path = os.path.join(self.context.dataset_checkpoint_path, sub_dir, "dataset_ckpt", f"rank_{self.context.rank}.mbt")
        self.checkpoint().save_to_file(path)
        
    def resume(self, dir):
        # if dir empty, return
        resume_dir = os.path.join(dir, "dataset_ckpt")
        if not os.path.exists(resume_dir):
            return
        if os.listdir(resume_dir) == []:
            return
        
        all_dataset_ckpt_dict: Dict[str, DatasetCheckpointList] = {}
        for path in os.listdir(resume_dir):
            abs_path = os.path.join(resume_dir, path)
            all_dataset_ckpt_dict[path] = DatasetCheckpointList.load_from_file(abs_path)
        
        final_ckpt = None
        first_rank_ckpt_path = f"rank_0.mbt"
        cur_rank_ckpt_name = f"rank_{self.context.rank}.mbt"
        
        # 1. check if world size change
        new_world_size = self.context.world_size
        old_world_size = all_dataset_ckpt_dict[first_rank_ckpt_path].world_size
        if new_world_size == old_world_size:
            final_ckpt = all_dataset_ckpt_dict[cur_rank_ckpt_name]
        else:
            logger.warning(f"World size changed from {old_world_size} to {new_world_size}, resuming from scratch")
            all_dataset_ckpt_list = [dataset_ckpt_list for dataset_ckpt_list in all_dataset_ckpt_dict.values()]
            final_ckpt = all_dataset_ckpt_list[0]
            for other_dataset_ckpt in all_dataset_ckpt_list[1:]:
                final_ckpt.merge(other_dataset_ckpt)
        
        # 2. check if dataset change
        new_dataset_cnt = len(self.weighted_dataset.datasets)
        old_dataset_cnt = len(all_dataset_ckpt_dict[first_rank_ckpt_path].checkpoint_list)
        if new_dataset_cnt == old_dataset_cnt and new_world_size == old_world_size:
            global_sample_index = 0
            global_current_samples_list = []
            for dataset_ckpt_list in all_dataset_ckpt_dict.values():
                global_sample_index += dataset_ckpt_list.sample_idx
                global_current_samples_list.append(dataset_ckpt_list.current_samples)
            global_current_samples = [sum(x) for x in zip(*global_current_samples_list)]
            final_ckpt.global_sample_idx = global_sample_index
            final_ckpt.global_current_samples = global_current_samples
        else:
            logger.warning("Dataset or world size change detected, reset sampler state")
            final_ckpt.global_sample_idx = 0
            final_ckpt.global_current_samples = [0] * new_dataset_cnt
            final_ckpt.sample_idx = 0
            final_ckpt.current_samples = [0] * new_dataset_cnt
            
        self.weighted_dataset.load_checkpoint(final_ckpt)

    
    def progress(self, global_checkpoint: List[DatasetCheckpointList]):
        # This is a approximate progress calculation
        # This is used after:
        #   global_ckpt_list = [_ for _ in range(world_size)]
        #   dist.all_gather_object(global_ckpt_list, dataloader.checkpoint())
        #   dataloader.progress(global_ckpt_list)
        # or other all gather functions

        merged_ckpt = None
        for ckpt in global_checkpoint:
            if merged_ckpt is None:
                merged_ckpt = ckpt
            else:
                merged_ckpt.merge(ckpt)
        
        path_len_map = self.weighted_dataset.get_path_len_map()
        progress_dict = {}

        for checkpoint in merged_ckpt.checkpoint_list:
            dataset_info = checkpoint.dataset_info
            used = checkpoint.used
            total_samples_per_epoch = path_len_map[dataset_info.path]
            consumed_samples = 0
            
            for chunk, index_set in used.active.items():
                consumed_samples += len(index_set)
            min_epoch = 2048 if used.done else 0
            for epoch, chunk_set in used.done.items():
                min_epoch = min(min_epoch, epoch)
                for chunk in chunk_set:
                    consumed_samples += (chunk.stop - chunk.start)
            if min_epoch > 0:
                consumed_samples += total_samples_per_epoch * min_epoch

            progress_percentage = consumed_samples / total_samples_per_epoch
            
            progress_dict[dataset_info.path] = {
                "samples_per_epoch": total_samples_per_epoch,
                "samples_consumed": consumed_samples,
                "progress": progress_percentage
            }
        return progress_dict


from modelbest_sdk.dataset.sampler.weighted_megatron_sampler import WeightedMegatronSampler
from modelbest_sdk.dataset.sampler.weighted_sampler import WeightedSampler

WEIGHTED_SAMPLER = 'weighted_sampler'
WEIGHTED_MEGATRON_SAMPLER = 'weighted_megatron_sampler'

class SamplerFactory:
    @staticmethod
    def create_sampler(sampler_type, weights, rank, world_size, seed=None, **kwargs):
        if sampler_type == WEIGHTED_SAMPLER:
            return WeightedSampler(weights, rank, world_size, seed, **kwargs)
        elif sampler_type == WEIGHTED_MEGATRON_SAMPLER:
            return WeightedMegatronSampler(weights, rank, world_size, seed, **kwargs)
        else:
            raise ValueError(f"Unsupported sampler type: {sampler_type}")


import numpy as np

from modelbest_sdk.dataset.sampler.sampler import Sampler


class WeightedSampler(Sampler):
    def __init__(self, weights, rank, world_size, seed, **kwargs):
        self.weights = weights
        self.n_samples = len(weights)
        np.random.seed(seed)
        
    def __call__(self):
        return np.random.choice(self.n_samples, p=self.weights)
    
    def remove_index(self, index):
        if 0 <= index < self.n_samples:
            # Remove the index by setting its weight to 0
            self.weights[index] = 0
            # Re-normalize the weights
            total_weight = np.sum(self.weights)
            if total_weight > 0:
                self.weights /= total_weight
            else:
                raise ValueError("All weights have been removed.")
        else:
            raise ValueError(f"Index {index} is out of bounds for a dataset with {self.n_samples} samples.")


from collections import defaultdict
import hashlib
import numpy as np
import torch

# from modelbest_sdk.dataset.collater.collater import Collater
from modelbest_sdk.dataset.batch_packer.batch_packer_factory import CPM_FLASH_ATTN_BATCH_PACKER, BatchPackerFactory
from modelbest_sdk.dataset.batch_packer.cpm_flash_attn_batch_packer import CpmFlashAttnBatchPacker
from modelbest_sdk.dataset.batch_packer.megatron_batch_packer import MegatronBatchPacker
from modelbest_sdk.dataset.sampler.sampler_factory import WEIGHTED_MEGATRON_SAMPLER, WEIGHTED_SAMPLER
from modelbest_sdk.dataset.segment.segment_factory import FIXED_LENGTH_SEGMENT
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc, DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class BatchedDataset(torch.utils.data.IterableDataset):
    def __init__(self, context, weighted_dataset: WeightedDataset, batch_size, max_len, batch_packer_type=CPM_FLASH_ATTN_BATCH_PACKER, drop_last=False, **kwargs):
        self.context = context
        self.weighted_dataset = weighted_dataset
        self.drop_last = drop_last
        self.dataset_cnt = len(self.weighted_dataset.dataset_info_list)
        self.batch_packer_type = batch_packer_type
        self.batch_size = batch_size
        self.max_len = max_len
        self.kwargs = kwargs
        

    def __iter__(self):
        self.batch_packer = BatchPackerFactory.create_batch_packer(self.batch_packer_type, self.batch_size, self.max_len, self.dataset_cnt, **self.kwargs)
        for data in self.weighted_dataset:
            data: DetailedDoc
            for batch in self.batch_packer(data):
                yield batch
                
        if not self.drop_last:
            for batch in self.batch_packer():
                yield batch
    
    def collate_fn(self, batch):
        return BatchPackerFactory.collate_fn(self.batch_packer_type)(batch)


if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)

    dataset_info_list = [
        DatasetInfo(
            path="human.BaseDoc.sstable",
            weight=1,
            max_epoch=1
        ),
        DatasetInfo(
            path="hot_chars.BaseDoc.sstable",
            weight=2,
            max_epoch=2
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list)
    
    w_dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list,
        segment_type=FIXED_LENGTH_SEGMENT,
        sampler_type=WEIGHTED_MEGATRON_SAMPLER,
        max_len=8192
    )
    
    dataset = BatchedDataset(
        context=context,
        weighted_dataset=w_dataset,
        batch_size=1,
        max_len=8192
    )
    
    for data in dataset:
        print(data)
        # print(data['indexes'].keys())
        # print(data['indexes'].values())
    


from collections import defaultdict
import hashlib
import torch


class BatchPacker:
    def __init__(self, batch_size: int, max_len: int, dataset_cnt: int, **kwargs):
        self.hash_to_tag = defaultdict()
    
    def encode_tags(self, token_ids_len, tag):
        tag_hash = self._get_tag_hash(tag)
        self.hash_to_tag[tag_hash] = tag
        return torch.full((token_ids_len,), tag_hash, dtype=torch.int64)
    
    def _get_tag_hash(self, tag: str) -> int:
        hash_obj = hashlib.sha256(tag.encode('utf-8'))
        return int(hash_obj.hexdigest(), 16) & ((1 << 63) - 1)        
            


from modelbest_sdk.dataset.segment.conditionl_fixed_length_segment import ConditionalFixedLengthSegment
from modelbest_sdk.dataset.segment.fixed_length_segment import FixedLengthSegment
from modelbest_sdk.dataset.segment.no_segment import NoSegment
from modelbest_sdk.dataset.segment.rolling_truncate_segment import RollingTruncateSegment

NO_SEGMENT = 'no_segment'
FIXED_LENGTH_SEGMENT = 'fixed_length_segment'
CONDITIONAL_FIXED_LENGTH_SEGMENT = 'conditional_fixed_length_segment'
ROLLING_TRUNCATE_SEGMENT = 'rolling_truncate_segment'
class SegmentFactory:
    @staticmethod
    def create_segment(segment_type, max_len: int, drop_last=False, **kwargs):
        segment_class = SegmentRegistry.get_segment(segment_type)
        return segment_class(max_len, drop_last, **kwargs)

class SegmentRegistry:
    _registry = {}

    @classmethod
    def register_segment(cls, key, segment_class):
        if key in cls._registry:
            raise ValueError(f"Segment type '{key}' is already registered.")
        cls._registry[key] = segment_class

    @classmethod
    def get_segment(cls, key):
        if key not in cls._registry:
            raise ValueError(f"Unsupported segment type: {key}")
        return cls._registry[key]

SegmentRegistry.register_segment(NO_SEGMENT, NoSegment)
SegmentRegistry.register_segment(FIXED_LENGTH_SEGMENT, FixedLengthSegment)
SegmentRegistry.register_segment(CONDITIONAL_FIXED_LENGTH_SEGMENT, ConditionalFixedLengthSegment)
SegmentRegistry.register_segment(ROLLING_TRUNCATE_SEGMENT, RollingTruncateSegment)


from collections import defaultdict, deque
import logging
from typing import Dict, Generator, List, Tuple

import numpy as np
import torch
from modelbest_sdk.dataset.batch_packer.batch_packer import BatchPacker
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk

logger = logging.getLogger(__name__)
class AudioBatchPacker(BatchPacker):
    def __init__(self, batch_size, max_len, dataset_cnt, **kwargs):
        super().__init__(batch_size, max_len, dataset_cnt, **kwargs)
        self.buffer = deque()
        self.current_length = 0
        self.max_total_length = batch_size * max_len
        self.batch_size = 1
        self.dataset_cnt = dataset_cnt

        
    def put(self, data: DetailedDoc):
        print(data.usage)
        self.buffer.append(data)
        self.current_length += data.mm_doc_seq.doc_seq[0].shape[1]
        
    def pop(self):
        lengths = []
        indexes: List[Tuple[int, Dict[Chunk, List[int]]]] = []
        tokens = torch.zeros((self.batch_size, 16, self.max_total_length), dtype=torch.int32)
        mask = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.float)
        position_ids = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        doc_ids = []
        # tags = torch.full((self.batch_size, self.max_total_length), dtype=torch.int64, fill_value=-1)

        span_begin = 0
        while self.buffer:
            data: DetailedDoc = self.buffer.popleft()
            position = data.position
            dataset_idx = data.dataset_idx
            mm_doc = data.mm_doc_seq.doc_seq[0]
            shape = mm_doc.shape
            token_info = torch.tensor(mm_doc.token_info).reshape(shape)
            # tag = mm_doc.tag[0]  # TODO: support multiple tags
            length = shape[1]
            span_end = span_begin + length
            
            tokens[0, :, span_begin:span_end] = token_info
            indexes.append((dataset_idx, data.indexes_dict))
            # tags[0, span_begin:span_end] = self.encode_tags(length, tag)
            position_ids[0, span_begin:span_end] = torch.from_numpy(np.arange(1, length + 1, dtype=np.int32))
            lengths.append(length)
            doc_ids.append(mm_doc.docid)
            span_begin = span_end
        
        mask[:, :] = 1.0
        
        cu_seqlens = torch.cat(
            [torch.tensor([0] + lengths).cumsum(dim=-1), torch.tensor([self.max_total_length], dtype=torch.int32)],
            dim=0,
        ).int()
        batch = {
            'tokens': tokens[:,:,:-1],
            'labels': tokens[:,:,1:],
            'loss_mask': mask[:, :-1],
            'position_ids': position_ids[:, :-1],
            'indexes': indexes,
            "cu_seqlens": cu_seqlens,
            "lengths": torch.tensor(sum(lengths)).int(),
            "max_seqlen": int(torch.max(cu_seqlens[1:] - cu_seqlens[:-1])),
            "doc_ids": doc_ids,
            # "tags": tags,
            # "hash_to_tag": self.hash_to_tag
        }
        self.current_length = 0
        return batch
        
    def will_exceed(self, data: DetailedDoc):
        if data is None:
            return False
        return self.current_length + data.mm_doc_seq.doc_seq[0].shape[1] > self.max_total_length

    def too_long(self, data: DetailedDoc):
        if data.mm_doc_seq.doc_seq[0].shape[1] > self.max_total_length:
            logger.warning(f"Document {data.mm_doc_seq.doc_seq[0].docid} is too long, length {data.mm_doc_seq.doc_seq[0].shape[1]} > {self.max_total_length}, truncate it.")
            return True
        return False
    
    def __call__(self, detailed_doc: DetailedDoc=None, pop_last=False) -> Generator[DetailedDoc, None, None]:
        if (pop_last and self.buffer):
            yield self.pop()
        if detailed_doc is not None:
            if self.too_long(detailed_doc):
                truncate_token_info = detailed_doc.mm_doc_seq.doc_seq[0].token_info[:self.max_total_length*16]
                detailed_doc.mm_doc_seq.doc_seq[0].token_info = truncate_token_info
                detailed_doc.mm_doc_seq.doc_seq[0].shape = [16, self.max_total_length]
            if self.will_exceed(detailed_doc):
                yield self.pop()
            self.put(detailed_doc)
        
    @staticmethod
    def collate_fn(batch):
        return batch[0]


import shutil
import unittest
import torch
from test.test_base import TestBase
from modelbest_sdk.dataset.batch_packer.batch_packer_factory import MEGATRON_BATCH_PACKER
from modelbest_sdk.dataset.sampler.sampler_factory import WEIGHTED_MEGATRON_SAMPLER
from modelbest_sdk.dataset.segment.segment_factory import FIXED_LENGTH_SEGMENT, CONDITIONAL_FIXED_LENGTH_SEGMENT
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetCheckpointList, DatasetInfo, DatasetInfoList
import os
import torch.distributed as dist
from modelbest_sdk.dataset.modelbest_dataloader import ModelbestDataloader
import torch.multiprocessing as mp

class TestCheckpoint(TestBase):
    def test_save_resume_with_same_settings(self):
        world_size = 2
        TestBase.generate_data(dataset_path='/tmp/text', file_count=1, doc_count_per_file=10, token_count=10, token_offset=1000, mask_pattern='text')
        TestBase.generate_data(dataset_path='/tmp/instruction', file_count=1, doc_count_per_file=10, token_count=10, token_offset=2000, mask_pattern='instruction')
        TestBase.generate_data(dataset_path='/tmp/other_text', file_count=1, doc_count_per_file=10, token_count=10, token_offset=1000, mask_pattern='text')
        mp.spawn(main, args=(world_size,), nprocs=world_size, join=True)
        new_world_size = 4
        mp.spawn(change_world_size, args=(new_world_size,), nprocs=new_world_size, join=True)
        shutil.rmtree('/tmp/text')
        shutil.rmtree('/tmp/instruction')
        shutil.rmtree('iter_10')
        shutil.rmtree('iter_20')
    
def setup(rank, world_size):
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'
    dist.init_process_group("gloo", rank=rank, world_size=world_size)

def cleanup():
    dist.destroy_process_group()

def main(rank, world_size):
    setup(rank, world_size)
    from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
    context = DatasetContext(
        rank=rank,
        world_size=world_size,
        dataset_checkpoint_path='',
        seed=rank+world_size,
    )
    dp_group = dist.new_group(ranks=list(range(world_size)))
    dataset_info_list = DatasetInfoList([
        DatasetInfo(
            path='/tmp/text',
            weight=1,
        ),
        DatasetInfo(
            path='/tmp/instruction',
            weight=2,
        ),
    ])
    dataloader = ModelbestDataloader(
        context, 
        dataset_info_list, 
        batch_size=2, 
        max_len=16, 
        segment_type=CONDITIONAL_FIXED_LENGTH_SEGMENT, 
        sampler_type=WEIGHTED_MEGATRON_SAMPLER, 
        batch_packer_type=MEGATRON_BATCH_PACKER,
        dp_group=dp_group
        )
    base_tokens = []
    base_position_ids = []
    base_indexes = []
    base_last_sample = []
    for i, batch in enumerate(dataloader, start=1):
        dataloader.update(batch["indexes"], batch["last_sample"])
        if i == 10:
            dataloader.save('iter_10')
        if i > 20 and i <= 30:
            base_tokens.append(batch['tokens'])
            base_position_ids.append(batch['position_ids'])
            base_indexes.append(batch['indexes'])
            base_last_sample.append(dict(batch['last_sample']))
        if i == 30:
            break
    dist.barrier()
    dataloader.resume('iter_10')
    for i, batch in enumerate(dataloader, start=11):
        dataloader.update(batch["indexes"], batch["last_sample"])
        if i == 20:
            dataloader.save('iter_20')
        if i > 20 and i <= 30:
            assert torch.equal(batch['tokens'], base_tokens[i-21])
            assert torch.equal(batch['position_ids'], base_position_ids[i-21])
            assert batch['indexes'] == base_indexes[i-21]
            assert dict(batch['last_sample']) == base_last_sample[i-21], f"{dict(batch['last_sample'])} != {base_last_sample[i-21]}"
        if i == 30:
            break

    dist.barrier()

    dataloader.resume('iter_20')
    for i, batch in enumerate(dataloader, start=21):
        dataloader.update(batch["indexes"], batch["last_sample"])
        assert torch.equal(batch['tokens'], base_tokens[i-21])
        assert torch.equal(batch['position_ids'], base_position_ids[i-21])
        assert batch['indexes'] == base_indexes[i-21]
        assert dict(batch['last_sample']) == base_last_sample[i-21]
        if i == 30:
            break
        
    new_dataset_info_list = DatasetInfoList([
        DatasetInfo(
            path='/tmp/text',
            weight=0,
        ),
        DatasetInfo(
            path='/tmp/instruction',
            weight=2,
        ),
        DatasetInfo(
            path='/tmp/other_text',
            weight=1,
        )
    ])
    new_dataloader = ModelbestDataloader(
        context, 
        new_dataset_info_list, 
        batch_size=2, 
        max_len=16, 
        segment_type=CONDITIONAL_FIXED_LENGTH_SEGMENT, 
        sampler_type=WEIGHTED_MEGATRON_SAMPLER, 
        batch_packer_type=MEGATRON_BATCH_PACKER,
        dp_group=dp_group
        )
    new_dataloader.resume('iter_20')
    for i, batch in enumerate(new_dataloader, start=21):
        new_dataloader.update(batch["indexes"], batch["last_sample"])
        # print(batch['indexes'])
        # print(batch['last_sample'])
        # no bug is good
        if i == 30:
            break
    cleanup()

def change_world_size(rank, world_size):
    setup(rank, world_size)
    from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
    context = DatasetContext(
        rank=rank,
        world_size=world_size,
        dataset_checkpoint_path='',
        seed=rank+world_size,
    )
    dp_group = dist.new_group(ranks=list(range(world_size)))
    dataset_info_list = DatasetInfoList([
        DatasetInfo(
            path='/tmp/text',
            weight=1,
        ),
        DatasetInfo(
            path='/tmp/instruction',
            weight=2,
        ),
    ])
    dataloader = ModelbestDataloader(
        context, 
        dataset_info_list, 
        batch_size=2, 
        max_len=16, 
        segment_type=CONDITIONAL_FIXED_LENGTH_SEGMENT, 
        sampler_type=WEIGHTED_MEGATRON_SAMPLER, 
        batch_packer_type=MEGATRON_BATCH_PACKER,
        dp_group=dp_group
        )
    dataloader.resume('iter_20')
    for i, batch in enumerate(dataloader, start=21):
        dataloader.update(batch["indexes"], batch["last_sample"])
        # print(batch['indexes'])
        # print(batch['last_sample'])
        # no bug is good
        if i == 30:
            break
    cleanup()
if __name__ == '__main__':
    unittest.main()


from typing import Tuple
import numpy as np
import concurrent.futures
import torch
from modelbest_sdk.dataset.prefetch_chunk_dataset import PrefetchChunkDataset
from modelbest_sdk.dataset.sampler.sampler_factory import WEIGHTED_SAMPLER, WEIGHTED_MEGATRON_SAMPLER,SamplerFactory
from modelbest_sdk.dataset.segment.segment_factory import FIXED_LENGTH_SEGMENT, NO_SEGMENT
from modelbest_sdk.dataset.segment_dataset import SegmentDataset
from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

logger = logging.getLogger(__name__)

class WeightedDataset(torch.utils.data.IterableDataset):
    def __init__(
        self, 
        context: DatasetContext, 
        dataset_info_list: DatasetInfoList,
        segment_type=NO_SEGMENT,
        sampler_type=WEIGHTED_SAMPLER,
        max_len=1024,
        prefetch_chunk_cnt=2,
        chunk_size=1024,
        dp_group=None,
        **kwargs
        ):
        
        self.context = context
        self.dataset_info_list = dataset_info_list.dataset_info_list

        self.dataset_weights = []
        self.datasets: List[SegmentDataset] = []
        self.prefetch_chunk_datasets: List[PrefetchChunkDataset] = []
        self.datasets_iter = []
        weights = []
        def process_dataset_info(dataset_info: DatasetInfo):
            if dataset_info.weight == 0:
                logger.warning(f"Dataset {dataset_info.path} has weight 0, won't sample, but will initialize it")
                # continue
            prefetch_chunk_dataset = PrefetchChunkDataset(
                context=context,
                dataset_info=dataset_info,
                prefetch_chunk_cnt=prefetch_chunk_cnt,
                chunk_size=chunk_size,
                dp_group=dp_group,
                **kwargs
            )
            dataset = SegmentDataset(
                prefetch_chunk_dataset=prefetch_chunk_dataset,
                segment_type=segment_type,
                max_len=max_len,
                **kwargs
            )
            return dataset, prefetch_chunk_dataset, dataset_info.weight

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = list(executor.map(process_dataset_info, self.dataset_info_list))

        for dataset, prefetch_chunk_dataset, weight in results:
            self.datasets.append(dataset)
            self.prefetch_chunk_datasets.append(prefetch_chunk_dataset)
            weights.append(weight)
        
        weights = np.array(weights)
        self.dataset_weights = weights / weights.sum()
        self.sampler_type = sampler_type

        
        self.current_samples = np.zeros(len(self.dataset_weights), dtype=np.int64)
        self.sample_idx = 0
        self.global_current_samples = np.zeros(len(self.dataset_weights), dtype=np.int64)
        self.global_sample_idx = 0
        
        self.kwargs = kwargs

        if dp_group is not None:
            self.init_broadcast(dp_group)

    def init_broadcast(self, dp_group):
        def get_broadcast_data(prefetch_chunk_dataset: PrefetchChunkDataset):
            prefetch_chunk_dataset.init_from_local()
            return prefetch_chunk_dataset.mbtable_partition.to_broadcast_data()
        
        broadcast_list = [0]*len(self.datasets)
        if self.context.rank == 0:
            start_time = time.time()
            with ThreadPoolExecutor() as executor:
                futures = {executor.submit(get_broadcast_data, dataset): i for i, dataset in enumerate(self.prefetch_chunk_datasets)}
                for future in as_completed(futures):
                    idx = futures[future]
                    try:
                        broadcast_list[idx] = future.result()
                    except Exception as e:
                        logger.error(f"Dataset initialization failed at index {idx} with error: {e}")
            elapsed_time = time.time() - start_time
            logger.debug(f"[{time.strftime('%Y-%m-%d %H:%M:%S')}] Initialized mbtable partitions. Time taken: {elapsed_time:.2f} seconds")
            logger.debug(f"[{time.strftime('%Y-%m-%d %H:%M:%S')}] Broadcasting mbtable partitions to other ranks: {broadcast_list}")
        broadcast_start_time = time.time()
        torch.distributed.broadcast_object_list(broadcast_list, src=0)
        broadcast_elapsed_time = time.time() - broadcast_start_time
        logger.debug(f"[{time.strftime('%Y-%m-%d %H:%M:%S')}] Broadcast completed. Time taken: {broadcast_elapsed_time:.2f} seconds")
        
        reconstruction_start_time = time.time()
        for i, dataset in enumerate(self.prefetch_chunk_datasets):
            dataset.init_from_broadcast(broadcast_list[i])
        reconstruction_elapsed_time = time.time() - reconstruction_start_time
        logger.debug(f"[{time.strftime('%Y-%m-%d %H:%M:%S')}] Reconstructed datasets from broadcast. Time taken: {reconstruction_elapsed_time:.2f} seconds")

        
    def __iter__(self):
        self.sampler = SamplerFactory.create_sampler(self.sampler_type, weights=self.dataset_weights, rank=self.context.rank, world_size=self.context.world_size, seed=self.context.seed, **self.kwargs)
        if self.sampler_type == WEIGHTED_MEGATRON_SAMPLER:
            self.sampler.resume(self.global_sample_idx, self.global_current_samples)
        for dataset in self.datasets:
            self.datasets_iter.append(iter(dataset))
        while True:
            if all(d.exhausted for d in self.prefetch_chunk_datasets):
                logger.warning(f"All dataset exhaust on rank {self.context.rank}")
                break
            idx = self.sampler()
            if self.prefetch_chunk_datasets[idx].exhausted:
                logger.warning(f"Dataset {idx} exhaust on rank {self.context.rank}")
                self.sampler.remove_index(idx)
                continue
            chosen_iter = self.datasets_iter[idx]
            try:
                data: DetailedDoc = next(chosen_iter)
                if data is None:
                    continue
                data.dataset_idx = idx
                yield data
            except StopIteration:
                continue

    def checkpoint(self):
        checkpoint_list = [dataset.checkpoint() for dataset in self.datasets]
        return DatasetCheckpointList(
            checkpoint_list=checkpoint_list,
            world_size=self.context.world_size,
            tp_size=self.context.tp_size,
            sample_idx=self.sample_idx,
            current_samples=self.current_samples.tolist()
        )
    
    def load_checkpoint(self, dataset_checkpoint_list: DatasetCheckpointList):
        for i, checkpoint in enumerate(dataset_checkpoint_list.checkpoint_list):
            self.datasets[i].load_checkpoint(checkpoint)
        if not self.kwargs.get('clear_sampler_state', False):
            self.sample_idx = dataset_checkpoint_list.sample_idx
            self.current_samples = np.array(dataset_checkpoint_list.current_samples)
            self.global_sample_idx = dataset_checkpoint_list.global_sample_idx
            self.global_current_samples = np.array(dataset_checkpoint_list.global_current_samples)
        
    def update(self, consumed_sample_indexes: List[Tuple[int, Dict[Chunk, List[int]]]], last_samples: Dict[int, LastSample]={}):
        '''
        Update dataset checkpoint with consumed samples and last samples in this batch.

        Args:
            consumed_samples: [(dataset_idx, samples), ...]
                - A list where each element represents a sampling action that contributes to a batch.
                - Each element is a tuple:
                    - dataset_idx: The index of the dataset from which samples are taken.
                    - samples: A dictionary where:
                        - key: The Chunk(epoch, start, stop) within the dataset.
                        - value: A list of indexes within the chunk.
                        - EXAMPLE: {Chunk(0, 0, 16): [0, 1, 2], Chunk(0, 16, 32): [16, 17, 18]}

            last_samples: {dataset_idx: LastSample, ...}
                - A dictionary mapping each dataset index to its last partially read sample.
                - LastSample represents the last sample that was not fully consumed in the batch.

                # NOTE: last_samples are used only in scenarios where the raw data is segmented 
                # into multiple lines. This ensures that during resumption, each batch can be 
                # exactly reproduced.
        '''
        for dataset_idx, indexes in consumed_sample_indexes:
            self.datasets[dataset_idx].update(indexes, last_samples.get(dataset_idx, None))
            self.sample_idx += 1
            self.current_samples[dataset_idx] += 1
            
    def __len__(self):
        return sum(len(dataset) for dataset in self.datasets)
    
    def get_path_len_map(self):
        return {dataset.path: len(dataset) for dataset in self.prefetch_chunk_datasets}

if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)

    dataset_info_list = [
        DatasetInfo(
            path="tmp.sstable",
            weight=1,
            max_epoch=1
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list=dataset_info_list)
    
    dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list,
        segment_type=NO_SEGMENT,
        sampler_type=WEIGHTED_SAMPLER,
    )

    for data in dataset:
        for mm_doc in data.mm_doc_seq.doc_seq:
            print(mm_doc.dtype)
            print(mm_doc.shape)
            print(np.array(mm_doc.token_info).shape)


from typing import Generator
from modelbest_sdk.dataset.segment.segment import Segment
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc


class NoSegment(Segment):
    def __call__(self, detailed_doc: DetailedDoc) -> Generator[DetailedDoc, None, None]:
        yield detailed_doc



from modelbest_sdk.dataset.batch_packer.audio_batch_packer import AudioBatchPacker
from modelbest_sdk.dataset.batch_packer.cpm_flash_attn_batch_packer import CpmFlashAttnBatchPacker
from modelbest_sdk.dataset.batch_packer.megatron_batch_packer import MegatronBatchPacker
from modelbest_sdk.dataset.batch_packer.mmseq_batch_packer import MmSeqBatchPacker


CPM_FLASH_ATTN_BATCH_PACKER = "cpm_flash_attn_batch_packer"
MEGATRON_BATCH_PACKER = "megatron_batch_packer"
AUDIO_BATCH_PACKER = "audio_batch_packer"
MMSEQ_BATCH_PACKER = "mmseq_batch_packer"

class BatchPackerFactory:
    @staticmethod
    def create_batch_packer(batch_packer_type, batch_size, max_len, dataset_cnt, **kwargs):
        packer_class = BatchPackerRegistry.get_batch_packer(batch_packer_type)
        return packer_class(batch_size, max_len, dataset_cnt, **kwargs)

    @staticmethod
    def collate_fn(batch_packer_type):
        return BatchPackerRegistry.get_collate_fn(batch_packer_type)

class BatchPackerRegistry:
    _registry = {}

    @classmethod
    def register_batch_packer(cls, key, packer_class):
        if key in cls._registry:
            raise ValueError(f"Batch packer type '{key}' is already registered.")
        cls._registry[key] = packer_class

    @classmethod
    def get_batch_packer(cls, key):
        if key not in cls._registry:
            raise ValueError(f"Unsupported batch packer type: {key}")
        return cls._registry[key]

    @classmethod
    def get_collate_fn(cls, key):
        packer_class = cls.get_batch_packer(key)
        return packer_class.collate_fn

BatchPackerRegistry.register_batch_packer(CPM_FLASH_ATTN_BATCH_PACKER, CpmFlashAttnBatchPacker)
BatchPackerRegistry.register_batch_packer(MEGATRON_BATCH_PACKER, MegatronBatchPacker)
BatchPackerRegistry.register_batch_packer(AUDIO_BATCH_PACKER, AudioBatchPacker)
BatchPackerRegistry.register_batch_packer(MMSEQ_BATCH_PACKER, MmSeqBatchPacker)


import thriftpy2
from thriftpy2.utils import deserialize, serialize

import os

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")
metadata_thrift = thriftpy2.load(os.path.join(proto_dir, "metadata.thrift"))

def enum_to_string(enum_value):
    if enum_value == metadata_thrift.TableType.RECORD_TABLE:
        return 'RECORD_TABLE'
    elif enum_value == metadata_thrift.TableType.LIST_TABLE:
        return 'LIST_TABLE'
    else:
        print('Unknown type: ', enum_value)
        return None

def str_to_enum(enum_class, enum_str):
    for name, value in enum_class.__dict__.items():
        if name == enum_str:
            return value
    return 0

class SSTableMetadata:
    def __init__(self,
                 table_type=None,
                 total_count=None,
                 tokenizer_version=None,
                 proto_type=None):
        self.table_type = table_type
        self.total_count = total_count
        self.tokenizer_version = tokenizer_version
        self.proto_type = proto_type
    
    @staticmethod
    def deserialize(bin):
        if bin == b'':
            return SSTableMetadata()
        return SSTableMetadata.from_thrift(
                deserialize(metadata_thrift.SSTableMetadata(), bin))

    @staticmethod
    def from_thrift(meta_proto):
        return SSTableMetadata(
            table_type=enum_to_string(meta_proto.tabletype),
            total_count=meta_proto.total_count,
            tokenizer_version=meta_proto.tokenizer_version,
            proto_type=meta_proto.proto_type,
        )

    def serialize(self):
        return serialize(self.to_thrift())
    
    def to_thrift(self):
        return metadata_thrift.SSTableMetadata(
            tabletype=str_to_enum(metadata_thrift.TableType, self.table_type),
            total_count=self.total_count,
            tokenizer_version=self.tokenizer_version,
            proto_type=self.proto_type,
        )
    
    def __repr__(self) -> str:
        return f"SSTableMetadata(table_type={self.table_type}, total_count={self.total_count}, tokenizer_version={self.tokenizer_version}, proto_type={self.proto_type})"




import ctypes
import logging
import os
import random
import time

class ByteArray(ctypes.Structure):
    _fields_ = [("data", ctypes.POINTER(ctypes.c_char)),
                ("length", ctypes.c_size_t)]

class ByteArrayList(ctypes.Structure):
    _fields_ = [("data_list", ctypes.POINTER(ByteArray)),
                ("size", ctypes.c_size_t)]

class MetaDataEntry(ctypes.Structure):
    _fields_ = [("key", ctypes.c_char_p),
                ("value", ctypes.c_char_p)]

class MetaDataList(ctypes.Structure):
    _fields_ = [("entries", ctypes.POINTER(MetaDataEntry)),
                ("count", ctypes.c_int)]

    def to_dict(self):
        # 将entries转换为Python字典
        return {self.entries[i].key.decode("utf-8"): self.entries[i].value.decode("utf-8") for i in range(self.count)}


lib_path = os.path.join(os.path.dirname(__file__), 'lib', 'libmbtable_sdk_shared.so')
lib = ctypes.CDLL(lib_path)


lib.MbTableOpen.argtypes = [ctypes.c_char_p]
lib.MbTableOpen.restype = ctypes.c_void_p

lib.MbTableClose.argtypes = [ctypes.c_void_p]
lib.MbTableClose.restype = None

lib.MbTableRead.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableRead.restype = ctypes.POINTER(ByteArray)

lib.MbTableReadList.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableReadList.restype = ctypes.POINTER(ByteArrayList)

lib.MbTableGetMetaData.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
lib.MbTableGetMetaData.restype = ctypes.POINTER(ByteArray)

lib.MbTableGetAllMetaData.argtypes = [ctypes.c_void_p, ctypes.POINTER(MetaDataList)]
lib.MbTableGetAllMetaData.restype = None

lib.MbTableGetEntryCount.argtypes = [ctypes.c_void_p]
lib.MbTableGetEntryCount.restype = ctypes.c_int32

# Get metadata by giving file path.
lib.GetFileMetaData.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
lib.GetFileMetaData.restype = ctypes.POINTER(ByteArray)

# Get entry count by giving file path.
lib.GetFileEntryCount.argtypes = [ctypes.c_char_p]
lib.GetFileEntryCount.restype = ctypes.c_int32

# Single record iterator methods
lib.MbTableCreateIterator.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableCreateIterator.restype = ctypes.c_void_p

lib.IteratorGetRecord.argtypes = [ctypes.c_void_p]
lib.IteratorGetRecord.restype = ctypes.POINTER(ByteArray)

lib.IteratorHasNext.argtypes = [ctypes.c_void_p]
lib.IteratorHasNext.restype = ctypes.c_bool

lib.IteratorNext.argtypes = [ctypes.c_void_p]
lib.IteratorNext.restype = None

lib.PrintIterKey.argtypes = [ctypes.c_void_p]
lib.PrintIterKey.restype = None

lib.IteratorDelete.argtypes = [ctypes.c_void_p]
lib.IteratorDelete.restype = None

# List record iterator methods
lib.MbTableCreateListIterator.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableCreateListIterator.restype = ctypes.c_void_p

lib.IteratorGetList.argtypes = [ctypes.c_void_p]
lib.IteratorGetList.restype = ctypes.POINTER(ByteArrayList)

lib.ListIteratorHasNext.argtypes = [ctypes.c_void_p]
lib.ListIteratorHasNext.restype = ctypes.c_bool

lib.ListIteratorNext.argtypes = [ctypes.c_void_p]
lib.ListIteratorNext.restype = None

lib.PrintListIterKey.argtypes = [ctypes.c_void_p]
lib.PrintListIterKey.restype = None

lib.ListIteratorDelete.argtypes = [ctypes.c_void_p]
lib.ListIteratorDelete.restype = None

lib.FreeByteArray.argtypes = [ctypes.POINTER(ByteArray)]
lib.FreeByteArray.restype = None

lib.FreeByteArrayList.argtypes = [ctypes.POINTER(ByteArrayList)]
lib.FreeByteArrayList.restype = None

logger = logging.getLogger(__name__)
class MbTable:
    def __init__(self, path: str, max_retry=5):
        self.path = path
        self.max_retry = max_retry
        self._len = None

    def get_handle(self):
        self.handle = lib.MbTableOpen(self.path.encode('utf-8'))
        retry = 0
        while not self.handle and retry < self.max_retry:
            assert_path_exists(self.path)
            retry += 1
            logger.error(f'Failed to open mbtable {self.path}, retrying in 5 second, retry times: {retry}')
            import time
            import random
            time.sleep(random.randint(1, 5))
            self.handle = lib.MbTableOpen(self.path.encode('utf-8'))
        if not self.handle:
            raise Exception(f'Failed to open mbtable {self.path} after {self.max_retry} retries')

        return self.handle

    def read(self, index):
        '''
        only use this method if you seek few non-sequential records, otherwise use iterator
        '''
        table_type = self.get_table_type()
        if table_type == b'':
            table_type = self.get_table_type_from_proto()
        if table_type == b'LIST_TABLE' or table_type == 'LIST_TABLE':
            raise TypeError('read() function does not support LIST_TABLE')

        self.get_handle()
        value_bytes = lib.MbTableRead(self.handle, index).contents
        value = ctypes.string_at(value_bytes.data, value_bytes.length)
        lib.FreeByteArray(value_bytes)
        self.close()
        return value

    def read_list(self, index):
        '''
        only use this method if you seek a record of list data, otherwise use
        iterator.
        '''
        table_type = self.get_table_type()
        if table_type == b'':
            table_type = self.get_table_type_from_proto()
        if table_type != b'LIST_TABLE' and table_type != 'LIST_TABLE':
            raise TypeError('readlist() function only supports LIST_TABLE')

        self.get_handle()
        ret = lib.MbTableReadList(self.handle, index).contents
        values = []
        for i in range(0, ret.size):
            values.append(ctypes.string_at(ret.data_list[i].data,
                                           ret.data_list[i].length))
        lib.FreeByteArrayList(ret)
        self.close()
        return values

    def get_table_type(self):
        return self.get_metadata('table_type')

    def get_table_type_from_proto(self):
        from modelbest_sdk.dataset.thrift_wrapper.metadata import SSTableMetadata
        data = self.get_file_metadata('metadata')
        if data != b'':
            meta = SSTableMetadata.deserialize(data)
            if meta.table_type is not None:
                return meta.table_type
        return b''
 
    def get_proto_type_from_proto(self):
        from modelbest_sdk.dataset.thrift_wrapper.metadata import SSTableMetadata
        data = self.get_file_metadata('metadata')
        if data != b'':
            meta = SSTableMetadata.deserialize(data)
            if meta.proto_type is not None:
                return meta.proto_type
        return b''
    
    def get_tokenizer_version_from_proto(self):
        from modelbest_sdk.dataset.thrift_wrapper.metadata import SSTableMetadata
        data = self.get_file_metadata('metadata')
        if data != b'':
            meta = SSTableMetadata.deserialize(data)
            if meta.tokenizer_version is not None:
                return meta.tokenizer_version
        return b''

    def get_metadata(self, metadata_key):
        self.get_handle()
        key_bytes = metadata_key.encode('utf-8')
        value_bytes = lib.MbTableGetMetaData(self.handle, key_bytes).contents
        value = ctypes.string_at(value_bytes.data, value_bytes.length)
        lib.FreeByteArray(value_bytes)
        self.close()
        return value
    
    def get_file_metadata(self, metadata_key):
        path_bytes = self.path.encode('utf-8')
        key_bytes = metadata_key.encode('utf-8')
        value_bytes = lib.GetFileMetaData(path_bytes, key_bytes).contents
        value = ctypes.string_at(value_bytes.data, value_bytes.length)
        lib.FreeByteArray(value_bytes)
        return value
        
    def get_all_metadata(self):
        self.get_handle()
        meta_list = MetaDataList()
        lib.MbTableGetAllMetaData(self.handle, ctypes.byref(meta_list))
        self.close()
        return meta_list.to_dict()

    def get_entry_count(self):
        self.get_handle()
        count = lib.MbTableGetEntryCount(self.handle)
        self.close()
        return count

    def get_entry_count_from_metadata(self):
        from modelbest_sdk.dataset.thrift_wrapper.metadata import SSTableMetadata
        data = self.get_file_metadata('metadata')
        if data != b'':
            meta = SSTableMetadata.deserialize(data)
            if meta.total_count is not None:
                return meta.total_count
        return -1

    def get_file_entry_count(self):
        path_bytes = self.path.encode('utf-8')
        count = lib.GetFileEntryCount(path_bytes)
        return count

    def __len__(self):
        if self._len is None:
            self._len = self.get_file_entry_count()
        return self._len

    def close(self):
        if self.handle:
            lib.MbTableClose(self.handle)
            self.handle = None
            
def assert_path_exists(path):
    retry = 0
    exist = os.path.exists(path)
    retry_on_not_exist = int(os.getenv('RETRY_ON_NOT_EXIST', 0))
    while not exist and retry < retry_on_not_exist:
        logger.error(f'{path} not exist, retrying in 5 second, retry times: {retry}')
        time.sleep(random.randint(1, 5))
        retry += 1
        exist = os.path.exists(path)
    assert exist, f'{path} not exist'
    
class MbTableIterator:
    def __init__(self, path, start_index=0, max_iterations=None):
        self.path = path
        self.start_index = start_index
        self.max_iterations = max_iterations
        self.current_iteration = 0
        self.iterator = None
        self.mbtable = None

    def __enter__(self):
        self.mbtable = MbTable(self.path)
        self.iterator = lib.MbTableCreateIterator(self.mbtable.get_handle(), self.start_index)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.iterator is not None:
            lib.IteratorDelete(self.iterator)
        if self.mbtable is not None:
            self.mbtable.close()

    def __iter__(self):
        return self

    def __next__(self):
        if self.max_iterations is not None and self.current_iteration >= self.max_iterations:
            raise StopIteration
        if lib.IteratorHasNext(self.iterator):
            self.current_iteration += 1
            value = self.get_record()
            lib.IteratorNext(self.iterator)
            return value
        else:
            raise StopIteration

    def next(self):
        # Python 2的兼容性方法
        return self.__next__()

    def get_record(self):
        ret = lib.IteratorGetRecord(self.iterator).contents
        record = ctypes.string_at(ret.data, ret.length)
        lib.FreeByteArray(ret)
        return record


class MbTableListIterator:
    def __init__(self, path, start_index=0, max_iterations=None):
        self.path = path
        self.start_index = start_index
        self.max_iterations = max_iterations
        self.current_iteration = 0
        self.iterator = None
        self.mbtable = None

    def __enter__(self):
        self.mbtable = MbTable(self.path)
        self.iterator = lib.MbTableCreateListIterator(
                self.mbtable.get_handle(), self.start_index)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.iterator:
            lib.ListIteratorDelete(self.iterator)
        if self.mbtable:
            self.mbtable.close()

    def __iter__(self):
        return self

    def __next__(self):
        if self.max_iterations is not None and self.current_iteration >= self.max_iterations:
            raise StopIteration
        if self.iterator and lib.ListIteratorHasNext(self.iterator):
            self.current_iteration += 1
            values = self.get_records()
            lib.ListIteratorNext(self.iterator)
            return values
        else:
            raise StopIteration

    def next(self):
        # Python 2的兼容性方法
        return self.__next__()

    def get_record(self):
        raise TypeError('get_record() is not supported, use get_records() instead.')

    def get_records(self):
        values = []
        if self.iterator:
            ret = lib.IteratorGetList(self.iterator).contents
            for i in range(0, ret.size):
                values.append(ctypes.string_at(ret.data_list[i].data,
                                               ret.data_list[i].length))
            lib.FreeByteArrayList(ret)
        return values

class TwinMbTable:
    def __init__(self, base_path: str, img_path: str=None):
        if img_path is None:
            self.base_path = base_path + '.sstable'
            self.img_path = base_path + '_img.sstable'
        else:
            self.base_path = base_path
            self.img_path = img_path
        self.initialize()

    def initialize(self):
        self.base_mbtable = MbTable(self.base_path)
        self.img_mbtable = MbTable(self.img_path)
        
        base_len = self.base_mbtable.get_entry_count_from_metadata()
        img_len = self.img_mbtable.get_entry_count_from_metadata()
        assert base_len == img_len, 'Base and image tables have different number of entries'
        self._len = base_len
    
    def __len__(self):
        return self._len
        

class TwinMbTableListIterator:
    def __init__(self, twin_mbtable: TwinMbTable, start_index=0, max_iterations=None):
        self.iterator1 = MbTableListIterator(twin_mbtable.base_path, start_index, max_iterations)
        self.iterator2 = MbTableListIterator(twin_mbtable.img_path, start_index, max_iterations)

    def __enter__(self):
        self.iterator1.__enter__()
        self.iterator2.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.iterator1.__exit__(exc_type, exc_val, exc_tb)
        self.iterator2.__exit__(exc_type, exc_val, exc_tb)

    def __iter__(self):
        return self

    def __next__(self):
        record1 = self.iterator1.__next__()
        record2 = self.iterator2.__next__()
        return record1, record2

if __name__ == '__main__':
    from modelbest_sdk.dataset.thrift_wrapper.base_doc import MmDocSeq, MmDoc, DocType, ImageDoc
    # # mbt_path = '/mnt/jfs-ee/jeeves-agi/projects/8691-qwen-megatron-sft-sst/checkpoints/sstable/whoru'
    # mbt_path = 'mm1.mbt'
    # mbtable = MbTable(mbt_path)
    # print(mbtable.get_entry_count())
    # # print(mbtable.read(1))
    # print(mbtable.get_metadata('meta_key'))
    # print(mbtable.get_all_metadata())
    # from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
    # with MbTableIterator(mbt_path) as it:
    #     for record in it:
    #         print(record)
    #         doc = MmDocSeq.deserialize(record)
    #         print(doc)
    #         break
        
    twin_mbtable = TwinMbTable('example/mbtable_data/mmdoc/test_obelics')
    with TwinMbTableListIterator(twin_mbtable) as it:
        for i, (base_entry, img_entry) in enumerate(it):
            print(f'Entry {i}:')
            mm_doc_seq = MmDocSeq.from_twin_list(base_entry, img_entry)
            print(mm_doc_seq)
            
            break



import os
import shutil
import unittest

from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetInfo, DatasetInfoList
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder


class TestBase(unittest.TestCase):
    def setUp(self):
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        self.data_dir = os.path.join(cur_dir, "mbtable_test_data")
        self.config_dir = os.path.join(cur_dir, "config")
        self.checkpoint_dir = os.path.join(cur_dir, "checkpoint")
        os.makedirs(self.data_dir, exist_ok=True)
        os.makedirs(self.config_dir, exist_ok=True)
        os.makedirs(self.checkpoint_dir, exist_ok=True)
        self.simple_dataset_config_path = os.path.join(self.config_dir , "simple_dataset_config.mbt")
        self.dist_dataset_config_path = os.path.join(self.config_dir , "dist_dataset_config.mbt")
        self.simple_dataset_checkpoint_path = os.path.join(self.checkpoint_dir, "simple_dataset_checkpoint")
        self.dist_dataset_checkpoint_path = os.path.join(self.checkpoint_dir, "dist_dataset_checkpoint")
        self.simple_dataset_context_path = os.path.join(self.config_dir, "simple_dataset_context.mbt")
        self.dist_dataset_context_path = os.path.join(self.config_dir, "dist_dataset_context.mbt")

        
        self.long_dataset_path = os.path.join(self.data_dir, "base_doc_long")
        self.short_dataset_path = os.path.join(self.data_dir, "base_doc_short")
        self.simple_dataset_info_list = DatasetInfoList([
            DatasetInfo(
                path=self.long_dataset_path,
                weight=1,
                max_epoch=1
            )  
        ])
        self.dist_dataset_info_list = DatasetInfoList([
            DatasetInfo(
                path=self.long_dataset_path,
                weight=1,
            ),
            DatasetInfo(
                path=self.short_dataset_path,
                weight=2
            )
        ])
        self.simple_dataset_info_list.save_to_file(self.simple_dataset_config_path)
        self.dist_dataset_info_list.save_to_file(self.dist_dataset_config_path)
        
        simple_dataset_context = DatasetContext(
            dataset_config_path=self.simple_dataset_config_path,
            dataset_checkpoint_path=self.simple_dataset_checkpoint_path,
        )
        dist_dataset_context = DatasetContext(
            rank=0,
            world_size=2,
            num_workers=1,
            dataset_config_path=self.dist_dataset_config_path,
            dataset_checkpoint_path=self.dist_dataset_checkpoint_path,
        )
        simple_dataset_context.save_to_file(self.simple_dataset_context_path)
        dist_dataset_context.save_to_file(self.dist_dataset_context_path)

        os.makedirs(self.long_dataset_path, exist_ok=True)
        for i in range(2):
            builder = MbTableBuilder(os.path.join(self.long_dataset_path, f"part_{i}.mbt"))
            for j in range(10):
                doc = BaseDoc()
                doc.docid = f"doc_{i}_{j}"
                doc.token_ids = [-1, 2, 3, 4, 5, 6, 0]
                doc.mask = [False, False, False, False, False, False, True]
                doc.tag=["long"]
                doc_str = doc.serialize()
                builder.write(doc_str)
            builder.flush()
        
        builder = MbTableBuilder(self.short_dataset_path)
        for j in range(20):
            doc = BaseDoc()
            doc.docid = f"doc_{j}"
            doc.token_ids = [-1, 2, 3, 0]
            doc.mask = [False, False, False, True]
            doc.tag=["short"]
            doc_str = doc.serialize()
            builder.write(doc_str)
        builder.flush()

    def tearDown(self):
        if os.path.exists(self.data_dir):
            shutil.rmtree(self.data_dir)
        if os.path.exists(self.config_dir):
            shutil.rmtree(self.config_dir)
        if os.path.exists(self.checkpoint_dir):
            shutil.rmtree(self.checkpoint_dir)

    @staticmethod    
    def generate_data(dataset_path: str, file_count: int=1, doc_count_per_file: int=10, token_count: int=10, token_offset: int=1000, mask_pattern: str="text"):
        os.makedirs(dataset_path, exist_ok=True)

        for i in range(file_count):
            builder = MbTableBuilder(os.path.join(dataset_path, f"part_{i}.mbt"))
            for j in range(doc_count_per_file):
                doc_id = f"doc_{i}_{j}"
                token_ids = [token_offset + k for k in range(token_count)]
                if mask_pattern == "text":
                    mask = [False] * (token_count - 1) + [True]
                elif mask_pattern == "instruction":
                    mid = (token_count + 1) // 2  # 计算中间位置
                    mask = [True] * (mid - 1) + [False] * (token_count - mid) + [True]
                else:
                    raise ValueError("Invalid mask pattern")
                doc = BaseDoc(token_ids=token_ids, mask=mask, docid=doc_id)
                doc_str = doc.serialize()
                builder.write(doc_str)
            builder.flush()


from collections import defaultdict
from typing import Dict, Generator, List, Tuple
import torch
from modelbest_sdk.dataset.batch_packer.batch_packer import BatchPacker
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc, Position
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk


class MegatronBatchPacker(BatchPacker):
    def __init__(self, batch_size, max_len, dataset_cnt, **kwargs):
        self.max_len = max_len
        self.attention_mask = self._create_attention_mask(max_len)
        self.dataset_cnt = dataset_cnt

    def _create_attention_mask(self, max_len):
        mask = torch.tril(torch.ones((max_len, max_len))).unsqueeze(0) < 0.5
        return mask

    def _create_position_ids(self, max_len_plus_one: int, positions: List[Position]) -> torch.Tensor:
        total_len = 0
        tensors = []
        
        for position in positions:
            length = position.length
            tensors.append(torch.arange(length))
            total_len += length
        
        assert total_len <= max_len_plus_one, f"total_len is {total_len}, but max_len_plus_one is {max_len_plus_one}"
        
        remaining_len = max_len_plus_one - total_len
        if remaining_len > 0:
            tensors.append(torch.arange(remaining_len))
        
        return torch.cat(tensors)


    def __call__(self, detailed_doc: DetailedDoc=None) -> Generator[DetailedDoc, None, None]:
        if detailed_doc is None:
            return
        tokens = torch.zeros(self.max_len + 1, dtype=torch.long)
        mask = torch.ones(self.max_len + 1, dtype=torch.bool) # default do not calculate loss for padding token
        actual_tokens = torch.Tensor(detailed_doc.base_doc.token_ids).long()
        actual_mask = torch.Tensor(detailed_doc.base_doc.mask).bool()
        tokens[:actual_tokens.shape[0]] = actual_tokens
        mask[:actual_mask.shape[0]] = actual_mask
        
        assert tokens.shape[0] == self.max_len + 1, f"tokens.shape[0] is {tokens.shape[0]}, but self.max_len + 1 is {self.max_len + 1}"
            
        position_ids = self._create_position_ids(self.max_len + 1, detailed_doc.positions)        
        
        indexes: Tuple[int, Dict[Chunk, List[int]]] = (detailed_doc.dataset_idx, detailed_doc.indexes_dict)
        last_sample = {detailed_doc.dataset_idx: detailed_doc.last_sample}
        yield {
            'tokens': tokens[:-1],
            'labels': tokens[1:],
            'loss_mask': (~mask[:-1]).float(),
            'attention_mask': self.attention_mask,
            'position_ids': position_ids[:-1],
            'indexes': indexes,
            'last_sample': last_sample,
            'dataset_id': torch.tensor(detailed_doc.dataset_idx)
        }
        
    def collate_fn(batch):
        batched_data = {k: [] for k in batch[0]}
        
        for item in batch:
            for key in item:
                batched_data[key].append(item[key])

        # 对于可以直接转换为张量的数据，使用 torch.stack
        for key in ['tokens', 'labels', 'loss_mask', 'attention_mask', 'position_ids', 'dataset_id']:
            batched_data[key] = torch.stack(batched_data[key])

        last_sample_aggregated_dict = defaultdict()

        for last_sample_dict in batched_data['last_sample']:
            for dataset_idx, last_sample in last_sample_dict.items():
                last_sample_aggregated_dict[dataset_idx] = last_sample # 取最后一个样本

        batched_data['last_sample'] = last_sample_aggregated_dict
        return batched_data


import torch
from modelbest_sdk.dataset.batch_packer.batch_packer_factory import CPM_FLASH_ATTN_BATCH_PACKER, MEGATRON_BATCH_PACKER
from modelbest_sdk.dataset.batched_dataset import BatchedDataset
from modelbest_sdk.dataset.modelbest_dataloader import ModelbestDataloader
from modelbest_sdk.dataset.segment.segment_factory import FIXED_LENGTH_SEGMENT
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetInfo, DatasetInfoList
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.dataset.weighted_dataset import WeightedDataset
from test.test_base import TestBase
import unittest


class TestBatchPacker(TestBase):
    def test_megatron_batch_packer(self):
        token_count = 10
        max_len = 16
        context = DatasetContext()
        TestBase.generate_data(dataset_path='/tmp/megatron_batch_packer', file_count=1, doc_count_per_file=10, token_count=token_count, token_offset=1000, mask_pattern='text')
        dataset_info_list = DatasetInfoList([DatasetInfo(path='/tmp/megatron_batch_packer', weight=1, max_epoch=1)])
        
        w_dataset = WeightedDataset(context, dataset_info_list, max_len=max_len, segment_type=FIXED_LENGTH_SEGMENT)
        dataset = BatchedDataset(context, weighted_dataset=w_dataset, batch_size=1, max_len=max_len, batch_packer_type=MEGATRON_BATCH_PACKER)
        start_idx = 0   
        expected_token_ids = list(range(1000, 1000+token_count))
        expected_mask = [False] * (token_count - 1) + [True]
        concatenated_token_ids = expected_token_ids * max_len
        concatenated_mask = expected_mask * max_len
        
        for batch in dataset:
            tokens = batch['tokens']
            labels = batch['labels']
            loss_mask = batch['loss_mask']
            position_ids = batch['position_ids']
            assert len(tokens) == max_len
            assert len(loss_mask) == max_len
            assert len(position_ids) == max_len
            assert len(labels) == max_len
            assert tokens.tolist() == concatenated_token_ids[start_idx:start_idx + max_len], f"Tokens do not match. Expected: {concatenated_token_ids[start_idx:start_idx + max_len]}, Got: {tokens}"
            assert labels.tolist() == concatenated_token_ids[start_idx + 1:start_idx + max_len + 1], f"Labels do not match. Expected: {concatenated_token_ids[start_idx + 1:start_idx + max_len + 1]}, Got: {labels}"
            assert (~(loss_mask.bool())).tolist() == concatenated_mask[start_idx:start_idx + max_len], f"Loss mask does not match. Expected: {concatenated_mask[start_idx:start_idx + max_len]}, Got: {loss_mask}"
            start_idx += max_len
            
            
if __name__ == '__main__':
    unittest.main()


import os
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

class CustomBuildPy(build_py):
    def run(self):
        # Generate git info
        os.system('python generate_git_info.py')
        # Continue with the regular build
        build_py.run(self)

setup(
    name='modelbest_sdk',
    version='0.2.5',
    author='HankyZhao',
    author_email='zhq980115@gmail.com',
    description='Everything about modelbest data include data format mbtable, dataset, dataloader, and tools',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://codeup.aliyun.com/64ddb0a87f62ff9b3d23ca15/modelbest_sdk',
    packages=find_packages(),
    cmdclass={'build_py': CustomBuildPy},
    package_data={'': ['git_info.txt']},
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)



from collections import defaultdict
import io
from PIL import Image
from typing import Dict, Generator, List, Tuple
import torch
from modelbest_sdk.dataset.batch_packer.batch_packer import BatchPacker
from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc, DocType, Position
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk


class MmSeqBatchPacker(BatchPacker):
    def __init__(self, batch_size, max_len, dataset_cnt, **kwargs):
        self.max_len = max_len
        self.dataset_cnt = dataset_cnt
        for k, v in kwargs.items():
            setattr(self, k, v)
            print(f"{k}={v}")


    def __call__(self, detailed_doc: DetailedDoc=None) -> Generator[DetailedDoc, None, None]:
        if detailed_doc is None:
            return
        print(f"usage: {detailed_doc.usage}")
        if detailed_doc.proto_type == BASE_DOC:
            print('we could pack base doc as well')
            print(detailed_doc.base_doc)
            return
        tokens = torch.zeros(self.max_len + 1, dtype=torch.long)
        mask = torch.ones(self.max_len + 1, dtype=torch.bool)
        position_ids = torch.zeros(self.max_len + 1, dtype=torch.long)
        img_idx = 0
        img_list = []
        img_bounds = []
        span_begin, span_end = 0, 0
        for mmdoc in detailed_doc.mm_doc_seq.doc_seq:
            if mmdoc.dtype == DocType.TXT:
                print(mmdoc.text)
                # txt_token = mmdoc.token_info
                # span_end = span_begin + len(txt_token)
                # tokens[span_begin:span_end] = torch.tensor(txt_token)
                mask[span_begin:span_end] = 0 # calculate loss 
                # print(mmdoc)
                # print(txt_token)
            elif mmdoc.dtype == DocType.IMG:
                image_size = mmdoc.image.size
                place_holder = get_placeholder(image_size=image_size, image_index=img_idx)
                img_idx += 1
                span_end = span_begin + len(place_holder)
                tokens[span_begin:span_end] = torch.tensor(place_holder)
                mask[span_begin:span_end] = 1 # not calculate loss 
                img_list.append(mmdoc.image)
                img_bounds.append((span_begin, span_end))
                # print(f"image_size={image_size}")
                # print(f"place_holder={place_holder}")
                # print(f"image={mmdoc.image}")
            span_begin = span_end

            
        position_ids[:span_end] = torch.arange(span_end)
        indexes: Tuple[int, Dict[Chunk, List[int]]] = (detailed_doc.dataset_idx, detailed_doc.indexes_dict)
        yield {
            'tokens': tokens[:-1],
            'labels': tokens[1:],
            'loss_mask': (~mask[:-1]).float(),
            'position_ids': position_ids[:-1],
            'image_list': img_list,
            'image_bounds': img_bounds,
            'indexes': indexes,
            'dataset_id': torch.tensor(detailed_doc.dataset_idx),
        }
        
    @staticmethod
    def collate_fn(batch):
        batched_data = {k: [] for k in batch[0]}
        
        for item in batch:
            for key in item:
                batched_data[key].append(item[key])

        # 对于可以直接转换为张量的数据，使用 torch.stack
        for key in ['tokens', 'labels', 'loss_mask', 'position_ids', 'dataset_id']:
            batched_data[key] = torch.stack(batched_data[key])
        return batched_data

import math

scale_resolution = 448
max_slice_nums = 9
query_len = 64
im_start_token_id = 128010
im_end_token_id = 128011
unk_token_id = 128002
slice_start_token_id = 128020
slice_end_token_id = 128021
im_id_start_token_id = 128022
im_id_end_token_id = 128023
new_schema = True
use_image_id = True


def get_image_placeholder(use_im_start_end=False):
    if use_im_start_end:
        return [im_start_token_id] + [unk_token_id] * query_len + [im_end_token_id]
    else:
        return [unk_token_id] * query_len


def get_slice_image_placeholder(use_im_start_end=False):
    if use_im_start_end:
        return [slice_start_token_id] + [unk_token_id] * query_len + [slice_end_token_id]
    else:
        return [unk_token_id] * query_len


_image_placeholder = get_image_placeholder(True)
_slice_placeholder = get_slice_image_placeholder(True)


def get_sliced_grid(image_size):
    original_width, original_height = image_size
    log_ratio = math.log(original_width / original_height)
    ratio = original_width * original_height / (scale_resolution * scale_resolution)
    multiple = min(math.ceil(ratio), max_slice_nums)
    if multiple <= 1:
        return None
    candidate_split_grids_nums = []
    for i in [multiple - 1, multiple, multiple + 1]:
        if i == 1 or i > max_slice_nums:
            continue
        candidate_split_grids_nums.append(i)
    
    candidate_grids = []
    for split_grids_nums in candidate_split_grids_nums:
        m = 1
        while m <= split_grids_nums:
            if split_grids_nums % m == 0:
                candidate_grids.append([m, split_grids_nums // m])
            m += 1

    best_grid = [1, 1]
    min_error = float("inf")
    for grid in candidate_grids:
        error = abs(log_ratio - math.log(grid[0] / grid[1]))
        if error < min_error:
            best_grid = grid
            min_error = error
    
    return best_grid


def get_grid_placeholder(grid):
    if grid is None:
        return []

    cols = grid[0]
    rows = grid[1]
    slices = []
    for i in range(rows):
        lines = []
        for j in range(cols):
            lines.append(_image_placeholder if not new_schema else _slice_placeholder)
        slices.append(sum(lines, []))
    
    if not new_schema:
        slice_placeholder = [slice_start_token_id] + sum(slices, []) + [slice_end_token_id]
    else:
        slice_placeholder = sum(slices, [])
    return slice_placeholder


def get_image_id_placeholder(idx):
    if use_image_id:
        return [im_id_start_token_id, idx, im_id_end_token_id]
    else:
        return []


def get_placeholder(image_size, image_index):
    grid = get_sliced_grid(image_size=image_size)
    return get_image_id_placeholder(image_index) \
        + _image_placeholder \
        + get_grid_placeholder(grid=grid)


import ctypes
import os

from modelbest_sdk.file_format.mbtable import ByteArray, ByteArrayList

lib_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'lib', 'libmbtable_sdk_shared.so'))
lib = ctypes.CDLL(lib_path)

lib.MbTableBuilderCreate.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
lib.MbTableBuilderCreate.restype = ctypes.c_void_p

lib.MbTableBuilderWrite.argtypes = [ctypes.c_void_p, ctypes.POINTER(ByteArray)]
lib.MbTableBuilderWrite.restype = ctypes.c_int

lib.MbTableBuilderWriteList.argtypes = [ctypes.c_void_p,
                                        ctypes.POINTER(ByteArrayList)]
lib.MbTableBuilderWriteList.restype = ctypes.c_int

lib.MbTableBuilderAddMetaData.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                          ctypes.POINTER(ByteArray)]
lib.MbTableBuilderAddMetaData.restype = None

lib.MbTableBuilderBuild.argtypes = [ctypes.c_void_p]
lib.MbTableBuilderBuild.restype = None

def create_byte_array(data):
    return ByteArray(ctypes.cast(ctypes.c_char_p(data), ctypes.POINTER(ctypes.c_char)), len(data))

class MbTableBuilder:
    def __init__(self, path: str, codec: str = 'kZlib'):
        """
        Initializes an instance of the class, setting up the underlying MbTable builder with the specified path and compression codec.

        The `codec` parameter specifies the compression method to be used for the MbTable. It must be a string that matches one of the predefined compression codec names. The available codecs are as follows:
        - 'kLzo': LZO compression.
        - 'kZlib': zlib compression.
        - 'kUnCompress': No compression.
        - 'kGzip': Gzip compression.
        - 'kSnappy': Snappy compression.
        - 'kUnknown': Represents an unknown or unsupported compression codec. It's recommended to use one of the supported codecs for compatibility.

        Args:
            path (str): The file system path where the MbTable will be created.
            codec (str, optional): The name of the compression codec to use. Defaults to 'kZlib'.

        Note:
            The path and codec arguments are encoded in UTF-8 before being passed to the underlying `CreateMbTableBuilder` function of the `lib` library.
        """
        dir_name = os.path.dirname(path)
        if dir_name == '':
            dir_name = '.'
        os.makedirs(dir_name, exist_ok=True)
        self.builder = lib.MbTableBuilderCreate(path.encode('utf-8'), codec.encode('utf-8'))

    def write(self, record):
        if isinstance(record, str):
            record = record.encode('utf-8')
        record_ba = create_byte_array(record)
        ret = lib.MbTableBuilderWrite(self.builder, ctypes.byref(record_ba))
        if ret == -1:
            raise Exception('Maximum of records 2^30 reached')
            
    def write_list(self, records):
        bal = ByteArrayList()
        bal.size = len(records)
        array_list = ByteArray * bal.size
        data_list = array_list()
        i = 0
        for record in records:
            if isinstance(record, str):
                record = record.encode('utf-8')
            data_list[i] = create_byte_array(record)
            i+=1
        bal.data_list = ctypes.cast(data_list, ctypes.POINTER(ByteArray))
        ret = lib.MbTableBuilderWriteList(self.builder, ctypes.byref(bal))
        if ret == -1:
            raise Exception('Maximum of records 2^20 reached')
            
    def add_metadata(self, key: str, value):
        if isinstance(value, str):
            value = value.encode('utf-8')
        value_ba =  create_byte_array(value)
        lib.MbTableBuilderAddMetaData(self.builder, key.encode('utf-8'), ctypes.byref(value_ba))

    def flush(self):
        lib.MbTableBuilderBuild(self.builder)

if __name__ == '__main__':
    from modelbest_sdk.dataset.thrift_wrapper.base_doc import MmDocSeq, MmDoc, DocType
    builder = MbTableBuilder('./mm1.mbt', 'kZlib')
    
    for i in range(10):
        doc_list = []
        for j in range(4):
            mmdoc = MmDoc(
                dtype=DocType.AUDIO,
                token_info=[1, 2, 3, 4, 5, 6],
                shape=[10, 10],
                mask=[False, False, True, True, False, False],
                docid='123',
                tag=['tag1', 'tag2'],
                version='1.0',
                reserved_col='reserved_col',
            )
            doc_list.append(mmdoc)
        doc_seq = MmDocSeq(doc_list)
        builder.write(doc_seq.serialize())
    builder.flush()



import io
import logging
from typing import Dict, List, Tuple, Union
import thriftpy2
from thriftpy2.utils import deserialize, serialize
import os
from PIL import Image
import torch
import numpy

from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk, LastSample
from modelbest_sdk.dataset.thrift_wrapper.utils import Utils

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")

doc_thrift = thriftpy2.load(os.path.join(proto_dir, "traindoc.thrift"))
mmdoc_thrift = thriftpy2.load(os.path.join(proto_dir, "mm_doc.thrift"))
imgdoc_thrift = thriftpy2.load(os.path.join(proto_dir, "image_doc.thrift"))

logger = logging.getLogger(__name__)

class BaseDoc:
    def __init__(self, token_ids=None, mask=None, docid=None, tag=None, token=None, tokenizer_version=None, reserved_col=None):
        self.token_ids = token_ids
        self.mask = mask
        self.docid = docid
        self.tag = tag
        self.token = token
        self.tokenizer_version = tokenizer_version
        self.reserved_col = reserved_col
    
    def split(self, offset, overlap=1) -> Tuple["BaseDoc", "BaseDoc"]:
        if not (1 <= offset <= len(self.token_ids)):
            raise ValueError("Offset must be between 1 and the length of the token_ids minus one.")
        # Split token_ids and mask with an overlap at offset
        token_ids_1, token_ids_2 = self.token_ids[:offset], self.token_ids[offset-overlap:]
        mask_1, mask_2 = self.mask[:offset], self.mask[offset-overlap:]
        
        doc1 = BaseDoc(token_ids=token_ids_1, mask=mask_1, docid=self.docid, tag=self.tag, token=self.token, tokenizer_version=self.tokenizer_version, reserved_col=self.reserved_col)
        doc2 = BaseDoc(token_ids=token_ids_2, mask=mask_2, docid=self.docid, tag=self.tag, token=self.token, tokenizer_version=self.tokenizer_version, reserved_col=self.reserved_col)
        return doc1, doc2
    
    def concat(self, other: "BaseDoc") -> "BaseDoc":
        if self.tag is None or other.tag is None:
            tag = None
        else:
            tag = list(set(self.tag + other.tag))
        return BaseDoc(
            token_ids=self.token_ids + other.token_ids,
            mask=self.mask + other.mask,
            docid=self.docid,
            tag=tag, # TODO：now there should be only one tag since concat happens in the same dataset, and same dataset should have only one tag
            token=self.token,
            tokenizer_version=self.tokenizer_version,
            reserved_col=self.reserved_col
        )

    @staticmethod
    def deserialize(bin):
        return BaseDoc.from_thrift(deserialize(doc_thrift.BaseDoc(), bin))
    
    def serialize(self):
        return serialize(self.to_thrift())
    
    def to_thrift(self):#反射
        return doc_thrift.BaseDoc(
            token_ids=self.token_ids,
            mask=self.mask,
            docid=self.docid,
            tag=self.tag,
            token=self.token,
            tokenizer_version=self.tokenizer_version,
            reserved_col=self.reserved_col
        )
    
    @staticmethod
    def from_thrift(thrift_base_doc):
        return BaseDoc(
            token_ids=thrift_base_doc.token_ids,
            mask=thrift_base_doc.mask,
            docid=thrift_base_doc.docid,
            tag=thrift_base_doc.tag,
            token=thrift_base_doc.token,
            tokenizer_version=thrift_base_doc.tokenizer_version,
            reserved_col=thrift_base_doc.reserved_col
        )
        
    def __len__(self) -> int:
        return len(self.token_ids)
        
    def __repr__(self) -> str:
        return f"BaseDoc(token_ids={self.token_ids}, mask={self.mask}, tag={self.tag}, token={self.token}, tokenizer_version={self.tokenizer_version}, reserved_col={self.reserved_col})"

from enum import Enum

class DocType(Enum):
    UNKNOWN = 0
    TXT = 1
    IMG = 2
    AUDIO = 3
    VIDEO = 4

class MmDoc:
    def __init__(self, dtype: DocType=None, token_info=None, text=None, shape=None, mask=None, docid=None, tag=None, version=None, reserved_col=None, image_md5=None):
        self.dtype = dtype
        self.token_info = token_info
        self.text = text
        self.shape = shape
        self.mask = mask
        self.docid = docid
        self.tag = tag
        self.version = version
        self.reserved_col = reserved_col
        self.image_md5 = image_md5
        self.image: Image.Image = None

    def split(self, offset, overlap=1) -> Tuple["MmDoc", "MmDoc"]:
        if not (1 <= offset <= self.shape[1]):
            raise ValueError("Offset must be between 1 and the length of the token_ids minus one.")
        # Split token_ids and mask with an overlap of one token at offset
        
        token_info_2d = torch.Tensor(self.token_info).reshape(self.shape)
        token_ids_1, token_ids_2 = token_info_2d[:, :offset].reshape(-1).tolist(), token_info_2d[:, offset-overlap:].reshape(-1).tolist()
        if self.mask is not None:
            mask_1, mask_2 = self.mask[:offset], self.mask[offset-overlap:]
        else:
            mask_1, mask_2 = None, None
        return MmDoc(
            dtype=self.dtype,
            token_info=token_ids_1,
            shape=[self.shape[0], offset],
            mask=mask_1,
            docid=self.docid,
            tag=self.tag,
            version=self.version,
            reserved_col=self.reserved_col
        ), MmDoc(
            dtype=self.dtype,
            token_info=token_ids_2,
            shape=[self.shape[0], self.shape[1] - offset + overlap],
            mask=mask_2,
            docid=self.docid,
            tag=self.tag,
            version=self.version,
            reserved_col=self.reserved_col
        )
        
    def concat(self, other: "MmDoc") -> "MmDoc":
        if self.tag is None or other.tag is None:
            tag = None
        else:
            tag = list(set(self.tag + other.tag))
        token_ids = torch.cat([torch.Tensor(self.token_info).reshape(self.shape), torch.Tensor(other.token_info).reshape(other.shape)], dim=1).reshape(-1).tolist()
        if self.mask is not None and other.mask is not None:
            mask = self.mask + other.mask
        else:
            mask = None
        return MmDoc(
            dtype=self.dtype,
            token_info=token_ids,
            shape=[self.shape[0], self.shape[1] + other.shape[1]],
            mask=mask,
            docid=self.docid,
            tag=tag,
            version=self.version,
            reserved_col=self.reserved_col
        )
        

    @staticmethod
    def from_thrift(thrift_mm_doc):
        return MmDoc(
            dtype=DocType(thrift_mm_doc.dtype),
            token_info=numpy.frombuffer(thrift_mm_doc.token_buffer, numpy.int32).tolist() if thrift_mm_doc.token_buffer is not None else thrift_mm_doc.token_info,
            text=thrift_mm_doc.text,
            shape=thrift_mm_doc.shape,
            mask=thrift_mm_doc.mask,
            docid=thrift_mm_doc.docid,
            tag=thrift_mm_doc.tag,
            version=thrift_mm_doc.version,
            reserved_col=thrift_mm_doc.reserved_col,
            image_md5=thrift_mm_doc.image_md5
        )

    def to_thrift(self):
        return mmdoc_thrift.MmDoc(
            dtype=self.dtype.value,
            token_info=self.token_info,
            text=self.text,
            shape=self.shape,
            mask=self.mask,
            docid=self.docid,
            tag=self.tag,
            version=self.version,
            reserved_col=self.reserved_col,
            image_md5=self.image_md5
        )
        
    def serialize(self):
        return serialize(self.to_thrift())
    
    @staticmethod
    def deserialize(bin):
        return MmDoc.from_thrift(deserialize(mmdoc_thrift.MmDoc(), bin))
    
    def __len__(self) -> int:
        if self.dtype == DocType.TXT:
            return len(self.token_info) if self.token_info else 0
        elif self.dtype == DocType.AUDIO:
            return self.shape[1]
        else:
            return 0
        # raise ValueError(f"unsupported dtype {self.dtype}")
    
    def __repr__(self) -> str:
        return f"MmDoc(dtype={self.dtype}, token_info={self.token_info}, text={self.text}, shape={self.shape}, mask={self.mask}, docid={self.docid}, tag={self.tag}, version={self.version}, reserved_col={self.reserved_col}, image_md5={self.image_md5})"


class Position:
    def __init__(self, chunk: Chunk, index: int, length: int=None, truncated: bool=False):
        self.chunk = chunk
        self.index = index
        self.length = length
        self.truncated = truncated
    
    def split(self, offset, overlap=1) -> Tuple["Position", "Position"]:
        if not (1 <= offset <= self.length):
            raise ValueError("Offset must be between 1 and the length of the position minus one.")
        # Split with an overlap of one token at offset
        position1 = Position(chunk=self.chunk, index=self.index, length=offset, truncated=True)
        position2 = Position(chunk=self.chunk, index=self.index, length=self.length - offset + overlap, truncated=True)
        return position1, position2

    def __repr__(self) -> str:
        return f"Position(chunk={self.chunk}, index={self.index}, length={self.length}, truncated={self.truncated})"

class MmDocSeq:
    def __init__(self, doc_seq: List[MmDoc]=None):
        self.doc_seq = doc_seq
        self.length = None
        
    @staticmethod
    def from_twin_list(base_list, image_list):
        mm_doc_seq = MmDocSeq.deserialize(base_list[0])
        image_list = [item for item in image_list if item != b'']
        image_inner_idx = 0
        for mm_doc in mm_doc_seq.doc_seq:
            if mm_doc.dtype == DocType.IMG:
                image_doc = ImageDoc.deserialize(image_list[image_inner_idx])
                image_bytes = image_doc.image_bytes
                img_io = io.BytesIO(image_bytes)
                img_io.seek(0)
                mm_doc.image = Image.open(img_io).convert('RGB')
                image_inner_idx += 1
                assert mm_doc.image_md5 == image_doc.image_md5, "image_md5 not match"
        return mm_doc_seq

    @staticmethod
    def from_thrift(thrift_mm_doc_seq):
        return MmDocSeq(doc_seq=[MmDoc.from_thrift(doc) for doc in thrift_mm_doc_seq.doc_seq])

    def to_thrift(self):
        return mmdoc_thrift.MmDocSeq(doc_seq=[doc.to_thrift() for doc in self.doc_seq])

    def serialize(self):
        return serialize(self.to_thrift())
    
    @staticmethod
    def deserialize(bin):
        return MmDocSeq.from_thrift(deserialize(mmdoc_thrift.MmDocSeq(), bin))
    
    def __len__(self) -> int:
        if self.length is None:
            self.length = sum(len(doc) for doc in self.doc_seq)
        return self.length
    
    def __repr__(self) -> str:
        return f"MmDocSeq(doc_seq={self.doc_seq})"

class ImageDoc:
    def __init__(self, image_bytes: bytes=None, image_md5: bytes=None):
        self.image_bytes = image_bytes
        self.image_md5 = image_md5
    
    @staticmethod
    def from_thrift(thrift_image_doc):
        return ImageDoc(image_bytes=thrift_image_doc.image_bytes, image_md5=thrift_image_doc.image_md5)
    
    def to_thrift(self):
        return imgdoc_thrift.ImageDoc(image_bytes=self.image_bytes, image_md5=self.image_md5)
    
    def serialize(self):
        return serialize(self.to_thrift())
    
    @staticmethod
    def deserialize(bin):
        return ImageDoc.from_thrift(deserialize(imgdoc_thrift.ImageDoc(), bin))
    
    def __repr__(self) -> str:
        return f"ImageDoc(image_bytes={self.image_bytes}, image_md5={self.image_md5})"


class DetailedDoc:
    def __init__(
        self, 
        proto_type: str=None, 
        base_doc: BaseDoc=None, 
        mm_doc_seq: MmDocSeq=None, 
        position: Position=None, 
        positions: List[Position]=None, 
        dataset_idx: int=None, 
        raw: str=None, 
        usage: str=None, 
        last_sample: LastSample=None, 
        prev_len: int=0
    ):
        self.proto_type = proto_type
        self.base_doc = base_doc
        self.mm_doc_seq = mm_doc_seq
        self.position = position
        self.positions = positions if positions is not None else [position]
        self.dataset_idx = dataset_idx
        self.raw = raw
        self.usage = usage
        self.last_sample = last_sample
        self.prev_len = prev_len
    
    @property
    def tag(self) -> Union[str, List[str]]:
        if self.base_doc is not None:
            return self.base_doc.tag
        else:
            return self.mm_doc_seq.doc_seq[0].tag
    
    @property
    def mask(self) -> List[int]:
        if self.base_doc is not None:
            return self.base_doc.mask
        else:
            return self.mm_doc_seq.doc_seq[0].mask
    
    @property
    def indexes_dict(self) -> Dict[Chunk, List[int]]:
        # there could be multiple positions for the same chunk, so we need to group them by chunk
        # also remove last_sample position if it exists
        indexes_dict = {}
        for pos in self.positions:
            if self.last_sample is not None and pos.chunk == self.last_sample.chunk and pos.index == self.last_sample.index:
                continue
            if pos.chunk not in indexes_dict:
                indexes_dict[pos.chunk] = []
            indexes_dict[pos.chunk].append(pos.index)
        return indexes_dict
    
    def split(self, offset, overlap=1) -> Tuple["DetailedDoc", "DetailedDoc"]:
        pos1, pos2 = self.position.split(offset, overlap)
        if self.base_doc is not None:
            doc1, doc2 = self.base_doc.split(offset, overlap)
            return DetailedDoc(proto_type=BASE_DOC, base_doc=doc1, position=pos1, dataset_idx=self.dataset_idx, raw=self.raw, usage=self.usage, prev_len=self.prev_len), \
                DetailedDoc(proto_type=BASE_DOC, base_doc=doc2, position=pos2, dataset_idx=self.dataset_idx, raw=self.raw, usage=self.usage, prev_len=self.prev_len+offset-overlap)
        else:
            doc1, doc2 = self.mm_doc_seq.doc_seq[0].split(offset, overlap)
            return DetailedDoc(proto_type=MM_DOC_SEQ, mm_doc_seq=MmDocSeq([doc1]), position=pos1, dataset_idx=self.dataset_idx, raw=self.raw, usage=self.usage, prev_len=self.prev_len), \
                DetailedDoc(proto_type=MM_DOC_SEQ, mm_doc_seq=MmDocSeq([doc2]), position=pos2, dataset_idx=self.dataset_idx, raw=self.raw, usage=self.usage, prev_len=self.prev_len+offset-overlap)

    @staticmethod
    def extend_position(self_positions: List[Position], other_positions: List[Position]):
        extended_positions = []
        extended_positions.extend(self_positions)
        extended_positions.extend(other_positions)
        return extended_positions
    
    def concat(self, other: "DetailedDoc") -> "DetailedDoc":
        # only used for merging data in the same dataset
        
        return DetailedDoc(
            proto_type=self.proto_type,
            base_doc=self.base_doc.concat(other.base_doc) if self.base_doc is not None else None,
            mm_doc_seq=MmDocSeq([self.mm_doc_seq.doc_seq[0].concat(other.mm_doc_seq.doc_seq[0])]) if self.mm_doc_seq is not None else None,
            positions=DetailedDoc.extend_position(self.positions, other.positions),
            dataset_idx=self.dataset_idx,
        )
        
    def deserialize(self) -> bool:
        try:
            if self.proto_type == BASE_DOC:
                self.base_doc = BaseDoc.deserialize(self.raw)
            elif self.proto_type == MM_DOC_SEQ:
                if isinstance(self.raw, bytes):
                    self.mm_doc_seq = MmDocSeq.deserialize(self.raw)
                else:
                    base_list, image_list = self.raw
                    self.mm_doc_seq = MmDocSeq.from_twin_list(base_list, image_list)
            self.position.length = len(self)
            self.raw = None
            return True
        except Exception as e:
            logger.error(f"Failed to deserialize detailed doc: {e}")
            return False
    
    def tag_or_default(self, default_tag: str):
        if self.proto_type == BASE_DOC:
            if self.base_doc.tag is None:
                self.base_doc.tag = default_tag
        elif self.proto_type == MM_DOC_SEQ: # TODO: in seq, tag is doc level not seq level, is this ok?
            if self.mm_doc_seq.doc_seq[0].tag is None:
                self.mm_doc_seq.doc_seq[0].tag = default_tag
    
    def __len__(self) -> int:
        if self.proto_type == BASE_DOC:
            return len(self.base_doc)
        elif self.proto_type == MM_DOC_SEQ:
            return len(self.mm_doc_seq)
    
    def __repr__(self) -> str:
        return f"DetailedDoc(base_doc={self.base_doc}, mm_doc={self.mm_doc_seq}, position={self.position}, dataset_idx={self.dataset_idx})"



from collections import defaultdict
import logging
from typing import Dict, List, Set
import thriftpy2
from thriftpy2.utils import deserialize, serialize
import os

from modelbest_sdk.dataset.constant import *
from modelbest_sdk.dataset.thrift_wrapper.utils import Utils

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")

dc_thrift = thriftpy2.load(os.path.join(proto_dir, "dataset_checkpoint.thrift"))
dc_thrift.Chunk.__hash__ = lambda self: hash((self.epoch, self.start, self.stop))

logger = logging.getLogger(__name__)


class DatasetCheckpointList:
    def __init__(self, checkpoint_list, world_size=None, tp_size=None, sample_idx=None, current_samples=None):
        self.checkpoint_list: List[DatasetCheckpoint] = checkpoint_list
        self.world_size = world_size
        self.tp_size = tp_size
        self.sample_idx = sample_idx # if world size change, or dataset cnt change, those are useless
        self.current_samples = current_samples
        self.global_sample_idx = 0
        self.global_current_samples = []
        
    @staticmethod
    def load_from_file(path):
        return DatasetCheckpointList.from_thrift(deserialize(dc_thrift.DatasetCheckpointList(), Utils.load_from_file(path)))

    def save_to_file(self, path):
        Utils.save_to_file(path, serialize(self.to_thrift()))

    def to_thrift(self):
        thrift_checkpoints = [checkpoint.to_thrift() for checkpoint in self.checkpoint_list]
        return dc_thrift.DatasetCheckpointList(
            checkpoint_list=thrift_checkpoints,
            world_size=self.world_size,
            tp_size=self.tp_size,
            sample_idx=self.sample_idx,
            current_samples=self.current_samples
        )

    @staticmethod
    def from_thrift(thrift_checkpoint_list):
        checkpoint = [DatasetCheckpoint.from_thrift(checkpoint) for checkpoint in thrift_checkpoint_list.checkpoint_list]
        return DatasetCheckpointList(
            checkpoint_list=checkpoint,
            world_size=thrift_checkpoint_list.world_size,
            tp_size=thrift_checkpoint_list.tp_size,
            sample_idx=thrift_checkpoint_list.sample_idx,
            current_samples=thrift_checkpoint_list.current_samples,
        )
    
    def merge(self, other: 'DatasetCheckpointList'):
        # Note that we assume that the two dataset checkpoint lists have the same length and the same order of datasets.
        # merge the dataset with same path
        for i in range(len(self.checkpoint_list)):
            self.checkpoint_list[i].merge(other.checkpoint_list[i])
            
    def __repr__(self) -> str:
        return f"DatasetCheckpointList(checkpoint_list={self.checkpoint_list}, world_size={self.world_size}, tp_size={self.tp_size}, sample_idx={self.sample_idx}, current_samples={self.current_samples})"
        


class DatasetCheckpoint:
    def __init__(self, dataset_info, used=None, chunk_size=None, num_chunks=None, last_sample=None):
        self.dataset_info: DatasetInfo = dataset_info
        self.used: Used = used
        self.chunk_size = chunk_size
        self.num_chunks = num_chunks
        self.last_sample: LastSample = last_sample
    
    def merge(self, other: 'DatasetCheckpoint'):
        assert self.dataset_info.path == other.dataset_info.path
        self.used.merge(other.used)

    def to_thrift(self):
        thrift_dataset_info = self.dataset_info.to_thrift() if self.dataset_info else None
        thrift_used = self.used.to_thrift() if self.used else None
        thrift_last_sample = self.last_sample.to_thrift() if self.last_sample else None
        return dc_thrift.DatasetCheckpoint(
            dataset_info=thrift_dataset_info,
            used=thrift_used,
            chunk_size=self.chunk_size,
            num_chunks=self.num_chunks,
            last_sample=thrift_last_sample
        )

    @staticmethod
    def from_thrift(thrift_checkpoint):
        dataset_info = DatasetInfo.from_thrift(thrift_checkpoint.dataset_info) if thrift_checkpoint.dataset_info else None
        used = Used.from_thrift(thrift_checkpoint.used) if thrift_checkpoint.used else None
        last_sample = LastSample.from_thrift(thrift_checkpoint.last_sample) if thrift_checkpoint.last_sample else None
        return DatasetCheckpoint(
            dataset_info=dataset_info,
            used=used,
            chunk_size=thrift_checkpoint.chunk_size,
            num_chunks=thrift_checkpoint.num_chunks,
            last_sample=last_sample
        )
        
    def __repr__(self) -> str:
        return f"DatasetCheckpoint(dataset_info={self.dataset_info}, used={self.used}, chunk_size={self.chunk_size}, num_chunks={self.num_chunks}, last_sample={self.last_sample})"


class DatasetInfoList:
    def __init__(self, dataset_info_list=None):
        self.dataset_info_list: List[DatasetInfo] = dataset_info_list if dataset_info_list is not None else []

    @staticmethod
    def load_from_file(path):
        return DatasetInfoList.from_thrift(deserialize(dc_thrift.DatasetInfoList(), Utils.load_from_file(path)))

    def save_to_file(self, path):
        Utils.save_to_file(path, serialize(self.to_thrift()))

    def to_thrift(self):
        return dc_thrift.DatasetInfoList(
            dataset_info_list=[info.to_thrift() for info in self.dataset_info_list]
        )

    @staticmethod
    def from_thrift(thrift_dataset_info_list):
        dataset_info_list = [DatasetInfo.from_thrift(info) for info in thrift_dataset_info_list.dataset_info_list]
        return DatasetInfoList(dataset_info_list=dataset_info_list)

    def __repr__(self) -> str:
        return f"DatasetInfoList(dataset_info_list={self.dataset_info_list})"

class DatasetInfo:
    def __init__(self, path, weight=None, max_epoch=None, proto_type=BASE_DOC, usage=None):
        self.path = path
        self.weight = weight
        self.max_epoch = max_epoch
        self.proto_type = proto_type
        self.usage = usage

    def to_thrift(self):
        return dc_thrift.DatasetInfo(
            path=self.path,
            weight=self.weight,
            max_epoch=self.max_epoch
        )

    @staticmethod
    def from_thrift(thrift_dataset_info):
        return DatasetInfo(
            path=thrift_dataset_info.path,
            weight=thrift_dataset_info.weight,
            max_epoch=thrift_dataset_info.max_epoch
        )
        
    def __repr__(self) -> str:
        return f"DatasetInfo(path={self.path}, weight={self.weight}, max_epoch={self.max_epoch})"


class Used:
    def __init__(self):
        self.active: Dict[Chunk, Set[int]] = {}
        self.done: Dict[int, Set[Chunk]] = {}
        self.epoch: int = 0
    
    def merge(self, other: 'Used'):
        for chunk, index_set in other.active.items():
            if chunk not in self.active:
                self.active[chunk] = set()
            self.active[chunk].update(index_set)
        
        for epoch, chunk_set in other.done.items():
            if epoch not in self.done:
                self.done[epoch] = set()
            self.done[epoch].update(chunk_set)
        
    def to_thrift(self):
        thrift_active = {chunk.to_thrift(): index_set for chunk, index_set in self.active.items()}
        thrift_done = {epoch: {chunk.to_thrift() for chunk in chunk_set} for epoch, chunk_set in self.done.items()}
        return dc_thrift.Used(active=thrift_active, done=thrift_done, epoch=self.epoch)

    @staticmethod
    def from_thrift(thrift_used):
        used = Used()
        used.epoch = thrift_used.epoch
        for thrift_chunk, index_set in thrift_used.active.items():
            used.active[Chunk.from_thrift(thrift_chunk)] = set(index_set)
        for epoch, thrift_chunk_set in thrift_used.done.items():
            used.done[epoch] = {Chunk.from_thrift(chunk) for chunk in thrift_chunk_set}
        return used
    
    def __eq__(self, other):
        if not isinstance(other, Used):
            return NotImplemented
        return (self.active == other.active and
                self.done == other.done and
                self.epoch == other.epoch)
        
    def __repr__(self) -> str:
        return f"Used(active={self.active}, done={self.done}, epoch={self.epoch})"


class Chunk:
    def __init__(self, epoch, start, stop):
        self.epoch = epoch
        self.start = start
        self.stop = stop

    def to_thrift(self):
        return dc_thrift.Chunk(epoch=self.epoch, start=self.start, stop=self.stop)

    @staticmethod
    def from_thrift(thrift_chunk):
        return Chunk(epoch=thrift_chunk.epoch, start=thrift_chunk.start, stop=thrift_chunk.stop)
    
    def __repr__(self) -> str:
        return f"Chunk(epoch={self.epoch}, start={self.start}, stop={self.stop})"

    def __hash__(self) -> int:
        return hash((self.epoch, self.start, self.stop))
    
    def __eq__(self, other: 'Chunk') -> bool:
        return self.epoch == other.epoch and self.start == other.start and self.stop == other.stop

class LastSample:
    def __init__(self, chunk: Chunk, index: int, offset: int):
        self.chunk = chunk
        self.index = index
        self.offset = offset
    
    def to_thrift(self):
        return dc_thrift.LastSample(
            chunk=self.chunk.to_thrift(),
            index=self.index,
            offset=self.offset
        )  
    
    @staticmethod
    def from_thrift(thrift_last_sample):
        return LastSample(
            chunk=Chunk.from_thrift(thrift_last_sample.chunk),
            index=thrift_last_sample.index,
            offset=thrift_last_sample.offset
        )  
    
    def __repr__(self) -> str:
        return f"LastSample(chunk={self.chunk}, index={self.index}, offset={self.offset})"
    
    def __eq__(self, other: 'LastSample') -> bool:
        return self.chunk == other.chunk and self.index == other.index and self.offset == other.offset
    
    def __hash__(self) -> int:
        return hash((self.chunk, self.index, self.offset))



from collections import deque
from typing import Generator
from modelbest_sdk.dataset.constant import MM_DOC_SEQ
from modelbest_sdk.dataset.segment.segment import Segment
from modelbest_sdk.dataset.thrift_wrapper.base_doc import DetailedDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import LastSample


class RollingTruncateSegment(Segment):
    def __init__(self, max_len: int, drop_last=False, **kwargs):
        self.max_len = max_len
        self.text_only_segment_length = self.audio_only_segment_length = self.max_len + 1
        if kwargs.get('text_only_segment_length') is not None:
            self.text_only_segment_length = kwargs['text_only_segment_length']
        if kwargs.get('audio_only_segment_length') is not None:
            self.audio_only_segment_length = kwargs['audio_only_segment_length']
        self.segment_length = None
        self.drop_last = drop_last
            
    def is_mix(self, detailed_doc: DetailedDoc) -> bool:
        if detailed_doc.proto_type == MM_DOC_SEQ:
            dtype_list = [mmdoc.dtype for mmdoc in detailed_doc.mm_doc_seq.doc_seq]
            if len(set(dtype_list)) > 1:
                return True
        return False
            
    def __call__(self, detailed_doc: DetailedDoc) -> Generator[DetailedDoc, None, None]:
        if self.is_mix(detailed_doc):
            yield detailed_doc
            return
        
        if self.segment_length is None:
            if detailed_doc.proto_type == MM_DOC_SEQ:
                self.segment_length = self.audio_only_segment_length
            else:
                self.segment_length = self.text_only_segment_length
                
        
        current_length = len(detailed_doc)
        
        while current_length > self.segment_length:
            data_to_yield, data_to_segment = detailed_doc.split(self.segment_length, overlap=0)
            # TODO: uncomment when resume excatly is needed
            lask_chunk = data_to_yield.position.chunk
            last_index = data_to_yield.position.index
            last_offset = data_to_yield.prev_len + self.segment_length
            data_to_yield.last_sample = LastSample(lask_chunk, last_index, last_offset)
            yield data_to_yield
            current_length = len(data_to_segment)
            detailed_doc = data_to_segment
        yield detailed_doc

